#-----------------------------------
# Imports
#-----------------------------------

from typing import Callable, Generic, Iterator, Type, TypeVar, List

from muphyn.packages.core.utils.log_manager import LogManager

#-----------------------------------
# Generic Type
#-----------------------------------

T = TypeVar('T', bound = Callable)

#-----------------------------------
# Class
#-----------------------------------

class ModelsEvent (Generic[T]) :
    """Est la classe qui permet de contenir des listeners de toute sorte."""

    # -------------
    # Contructors
    # -------------

    def __init__(self) -> None:
        super().__init__()

        self._list : List[T] = []

    # -------------
    # Methods
    # -------------

    def __iter__ (self) -> Iterator[Type[T]] :
        """Permet de récuperer un itérateur pour boucler sur tous les éléments du tableau."""
        return self._list.__iter__()

    def __add__ (self, value) -> None :
        """Permet d'ajouter un événement."""

        if value is None :
            return
            
        if isinstance(value, T.__bound__) :
            self._list.append(value)

    def __sub__ (self, value) -> None :
        """Permet de supprimer un événement."""

        if value is None : 
            return

        if isinstance(value, type(T)) :
            self._list.remove(value)

    def clear (self) -> None :
        """Permet de supprimer tous les événements."""
        self._list.clear()

    def __len__ (self) -> int : 
        """Permet de récuperer la taille de la liste des événements."""
        return self._list.__len__()

    def __getitem__ (self, key : int) -> Type[T] :
        """Permet de récuperer l'événement à l'index demandé."""
        return self._list.__getitem__(key)
    
    def __setitem__ (self, key : int, value : Type[T]) -> None :
        """Empèche une erreur et ne fait rien en cas de tentative d'appel de la méthode."""
        LogManager().error('model event set item does nothing !!!')

    def __contains__ (self, item : Type[T]) -> bool :
        """Permet de savoir si l'événement se trouve déjà dans la liste."""
        return self._list.__contains__(item)


#-----------------------------------
# Imports
#-----------------------------------

# General Imports
import typing

# PyQt5 Imports
from PyQt5 import QtGui
from PyQt5.QtCore import QPointF, QSizeF
from PyQt5.QtWidgets import QGraphicsItem, QStyleOptionGraphicsItem, QWidget

# Project Imports
from muphyn.packages.interface.models.graphical_models.abstract_moveable_graphical_element import AbstractMoveableGraphicalElement

#-----------------------------------
# Class
#-----------------------------------

class TextModel (AbstractMoveableGraphicalElement) :
    """Est le modèle pour afficher du texte à l'écran."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, name : str, position : QPointF, size : QSizeF, rotation : float = 0.0, text : str = '', parent : QGraphicsItem = None) :
        
        AbstractMoveableGraphicalElement.__init__(self, name, position, size, rotation, text, parent)
    
    # -------------
    # Methods
    # -------------

    def paint (self, painter: QtGui.QPainter, option: QStyleOptionGraphicsItem, widget: typing.Optional[QWidget] = ...) -> None :
        painter.drawText(self.position, self.rendered_text)



from muphyn.packages.core.box_library.plci_core_boxes_libraries import *

from muphyn.packages.unit_test.unit_test import get_boxes_libraries

boxesLibraries = get_boxes_libraries()
box = boxesLibraries.construct_box('Boxes.Sources', 'Step')
print(box)



from ..core.plci_core_scheduler import Scheduler
from ..core.plci_core_data_type import DataType
from ..core.plci_core_signal import Signal
from ..core.plci_core_scheduler_params import SchedulerParams

from .unit_test import get_boxes_libraries, get_schedulers_library, simulate

# Get library
boxesLibraries = get_boxes_libraries()
schedulersLibraries = get_schedulers_library()

print("Creation of the diagram")
# Scheduler parameters
scheduler : Scheduler = schedulersLibraries.construct_scheduler('Schedulers', 'Default')
scheduler.params = SchedulerParams(10.0, 0.001)

# Signals
scheduler.diagram.append(Signal(index_ = 0, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 1, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 2, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 3, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 4, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 5, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 6, signal_type_ = DataType.FLOAT, default_value_ = 0.0))

# Sources
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sources', 'Ramp', start_time = 0.0, initial_value = 0.0, slope = 1.0))

# Constants
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sources', 'Constant', value = -5.0))

# Additioner
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Math', 'Addition'))

# Multiplier
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Math', 'Multiplier'))

# Integrator
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Math', 'Integrator'))

# Derivator
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Math', 'Derivator'))

# Sink
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sinks', 'Graph', start_time = 0.0, stop_time = scheduler.stop_time, title = "x_1"))
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sinks', 'Graph', start_time = 0.0, stop_time = scheduler.stop_time, title = "x_squared"))
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sinks', 'Graph', start_time = 0.0, stop_time = scheduler.stop_time, title = "integrated_x_squared"))
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sinks', 'Graph', start_time = 0.0, stop_time = scheduler.stop_time, title = "derivated_x_squared"))
 
# Diagram creation
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[0], scheduler.diagram.signals[0])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[1], scheduler.diagram.signals[1])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[2], scheduler.diagram.signals[2], scheduler.diagram.signals[6])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[3], scheduler.diagram.signals[3])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[4], scheduler.diagram.signals[4])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[5], scheduler.diagram.signals[5])

scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[2], scheduler.diagram.signals[0], scheduler.diagram.signals[1])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[3], scheduler.diagram.signals[2], scheduler.diagram.signals[6])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[6], scheduler.diagram.signals[6])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[7], scheduler.diagram.signals[3])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[4], scheduler.diagram.signals[3])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[5], scheduler.diagram.signals[3])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[8], scheduler.diagram.signals[4])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[9], scheduler.diagram.signals[5])

# Launch solver
simulate(scheduler)


#-----------------------------------
# Imports
#-----------------------------------

from typing import Callable, List, Any

#-----------------------------------
# Class
#-----------------------------------

class ModelSignal : 
    """Est la classe permettant de gérer des listeners de manière abstraite."""

    # -------------
    # Contructors
    # -------------

    def __init__ (self) : 
        self._slots : List[Callable] = []

    # -------------
    # Methods
    # -------------

    def emit (self, **kwargs : Any) :
        """Permet d'appeler chacun des slots avec les arguments passés."""

        for slot in self._slots : 
            slot(kwargs)

    def connect (self, slot_ : Callable) :
        """Permet d'ajouter le slot à la liste des slots."""

        if slot_ is None : 
            return

        if not(slot_ in self._slots) : 
            self._slots.append(slot_) 

    def disconnect (self, slot_ : Callable) :
        """Permet de supprimer le slot de la liste des slots."""

        if slot_ in self._slots :
            self._slots.remove(slot_)
    


# General import
import os

# PyQt import
from PyQt5.QtCore import QPointF, QRectF, QSizeF
from PyQt5.QtWidgets import QGraphicsItem

# Project import
from muphyn.packages.interface.shapes.basic_shape import GroupedShapes, ImageShapeFactory

class ComponentConnector(GroupedShapes):

    # Constants
    ItemSideSize = 8

    def __init__(self, group_position: QPointF = QPointF, parent: QGraphicsItem = None) -> None:
        super().__init__(group_position, parent)

    def absoluteCenter(self) -> QPointF:
        return self.scenePos() + self.boundingRect().center()

class AbstractComponentModel(GroupedShapes):

    # -------------
    # Constants
    # -------------
    DefaultSize: QSizeF = QSizeF(80, 80)
    DefaultRect: QRectF = QRectF(
        -ComponentConnector.ItemSideSize, 
        -ComponentConnector.ItemSideSize,
        DefaultSize.width() + 2*ComponentConnector.ItemSideSize,
        DefaultSize.height() + 2*ComponentConnector.ItemSideSize
    )

    def __init__(self, iconPath: str = None, group_position = QPointF(), parent = None) -> None:

        if iconPath is None or not os.path.exists(iconPath):
            raise(FileExistsError(f"AbstractComponentModel.__init__(): Given Icon path doesn't exists {iconPath}"))

        super().__init__(group_position, parent)

        # Icon
        self.icon = ImageShapeFactory(iconPath, size=AbstractComponentModel.DefaultSize, parent=self)

        # Set transform origin point to the center of this item
        self.setTransformOriginPoint(AbstractComponentModel.DefaultRect.center())

    # -------------
    # Properties
    # -------------

    

    # -------------
    # Methods
    # -------------
    def boundingRect(self) -> QRectF:
        return AbstractComponentModel.DefaultRect


# -----------------------------------
# Imports
# -----------------------------------

from typing import List
from scipy.integrate import solve_ivp
from scipy.signal import lsim, StateSpace
import ast
import numpy as np

# -----------------------------------
# Functions
# -----------------------------------

def get_coeff_vector(coeff_vector_string: str) -> List[float]:

    if coeff_vector_string is None:
        return [0]

    vector = []
    for sub in coeff_vector_string.split(' '):

        sub = sub.strip()
        if len(sub) == 0:
            continue

        try:
            vector.append(float(sub))
        except ValueError:
            ...

    return vector


def _init_transfer_function(box, simulation_params) -> None:
    
    box['A_list'] = box['A']
    box['B_list'] = box['B']
    box['C_list'] = box['C']
    box['D_list'] = box['D']

    box['system'] = StateSpace(box['A'], box['B'], box['C'], box['D']).to_tf()

    box['input_size'] = 1 #np.array(box['D_list']).shape[1]

    box['last_X'] = ast.literal_eval(box['initial_state'])
    
    box['last_timing'] = -1
    box['last_input'] = 0

def _function_transfer_function(box, event_) -> List :             

    if box['last_timing'] == event_.timing:    # if we are still on the current step, leave
        return []

    events: List = []

    step = event_.step_time

    t0 = event_.timing

    t = [t0, t0 + step]                 # time vector is only current and next step

    sol = lsim(box['system'], [box['last_input'], box.get_input(0).value], t, box['last_X'])
    box['last_input'] = box.get_input(0).value
    
    print(sol)
    out = sol[1][1]               # use the last values as the output of the step
    print(out)
    box['last_X'] = sol[2][-1]

    for output in box.outputs :
        events.append(box.construct_signal_event(output, out)) # only send the y(t) output, discard eventual derivatives for order >= 2

    box['last_timing'] = event_.timing

    return events



#-----------------------------------
# Imports
#-----------------------------------

# General Imports
from enum import Enum
from typing import Iterable, List, Any, Union

# PyQt5 Imports
from PyQt5.QtCore import QPointF, QMarginsF, QRectF, QSize, QSizeF, Qt, pyqtSignal
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QApplication, QGraphicsItem, QGraphicsSceneHoverEvent, QGraphicsSceneMouseEvent

# Project Imports
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.interface.models.event_signal_data import EventSignalData
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement
from muphyn.packages.interface.models.graphical_models.abstract_moveable_graphical_element import AbstractMoveableGraphicalElement
from muphyn.packages.interface.models.signals_model.abstract_connection_model import AbstractConnectionModel
from muphyn.packages.interface.models.signals_model.input_connection_model import InputConnectionModel, InputConnectionGroupModel
from muphyn.packages.interface.models.signals_model.output_connection_model import OutputConnectionModel, OutputConnectionGroupModel
from muphyn.packages.interface.models.signals_model.signal_link_model import SignalLinkModel
from muphyn.packages.interface.shapes.basic_shape import Rectangle, Circle, Polygon, Text
from muphyn.packages.interface.utils.graphical_items import BoxModelInfoArea
from muphyn.packages.interface.utils.interface_constants import MuphynFonts

#-----------------------------------
# Class
#-----------------------------------

io_margin : int = 4
name_margin : QPointF = QPointF(10, 5)

class BoxType(Enum):
    Math = 1
    Signal = 2
    Source = 3

class AbstractBoxModel (AbstractMoveableGraphicalElement) :
    """Est la classe abstraite des éléments capable d'être rendu sous forme de boxes et affichés à l'écran."""

    # -------------
    # Constants
    # -------------
    MinimunBoxWidth = 80
    MinimunBoxHeight = 80

    # -------------
    # Signals
    # -------------    
    input_count_changed = pyqtSignal()
    output_count_changed = pyqtSignal()

    # -------------
    # Constructors
    # -------------

    def __init__ (self, name : str, position : QPointF, size : QSizeF, rotation : float = 0.0, 
                    text : str = '', box_type: BoxType = BoxType.Signal, icon: str = None, parent : QGraphicsItem = None) :
        self._font_metrics_used = None

        AbstractMoveableGraphicalElement.__init__(self, name, position, size, rotation, text, parent)

        # Inputs
        self._inputs : List[InputConnectionModel] = []
        self._inputs_groups: dict[str, InputConnectionGroupModel] = {}

        # Outputs
        self._outputs : List[OutputConnectionModel]  = []
        self._outputs_groups: dict[str, OutputConnectionGroupModel] = {}

        self._size_text : QSize = QSize(0, 0)

        self._box_type = box_type

        if box_type == BoxType.Math:
            self.box_body: Rectangle = Rectangle(QPointF(InputConnectionModel.ItemWidth, 0), 
                AbstractBoxModel.MinimunBoxWidth, AbstractBoxModel.MinimunBoxHeight, border_color=Qt.GlobalColor.black, fill_color=Qt.GlobalColor.white, parent=self)
        elif box_type == BoxType.Signal:
            self.box_body: Polygon = Polygon( 
                [
                    QPointF(InputConnectionModel.ItemWidth, 0), 
                    QPointF(InputConnectionModel.ItemWidth, AbstractBoxModel.MinimunBoxHeight), 
                    QPointF(InputConnectionModel.ItemWidth + AbstractBoxModel.MinimunBoxWidth, AbstractBoxModel.MinimunBoxHeight/2)
                ],
                border_color=Qt.GlobalColor.black, 
                fill_color=Qt.GlobalColor.white,
                parent=self
                )
        elif box_type == BoxType.Source:
            radius = min(AbstractBoxModel.MinimunBoxWidth, AbstractBoxModel.MinimunBoxHeight) / 2
            center = QPointF(AbstractBoxModel.MinimunBoxWidth/2, AbstractBoxModel.MinimunBoxHeight/2) + QPointF(InputConnectionModel.ItemWidth, 0)
            self.box_body: Circle = Circle(center, radius, 
                border_color=Qt.GlobalColor.black, fill_color=Qt.GlobalColor.white, parent=self)
        else:
            raise(f"AbstractBoxModel.__init__(): Box Type is not supported: {box_type}")

        
        # Name 
        box_name_position = self.box_body.boundingRect().bottomLeft() + self.box_body.pos()
        self.name_label: Text = Text(name, box_name_position, text_max_width=AbstractBoxModel.MinimunBoxWidth, 
            font=MuphynFonts.BoxModelDetailsFont, alignment=Qt.AlignmentFlag.AlignHCenter, parent=self)

        # Box Body hiver event
        self.box_body.setAcceptHoverEvents(True)        
        self.box_body.hoverEnterEvent = self.boxBodyHoverEnterEvent
        self.box_body.hoverLeaveEvent = self.boxBodyHoverLeaveEvent

        # Box Details
        margins = 5
        if box_type == BoxType.Source:
            margins = 15
        elif box_type == BoxType.Signal:
            margins = QMarginsF(0, 20, 40, 20)
        self.box_details = BoxModelInfoArea(self.box_body.boundingRect().size(), self.name, image_path=icon, value=None, margins=margins, parent=self.box_body)
    
        # Init UI update handler
        self._should_recompute_inputs = False
        self._should_recompute_outputs = False
        self._should_recompute_box_details = False

    # -------------
    # Properties
    # -------------
    @property
    def inputs (self) -> Iterable[InputConnectionModel] :
        """Permet de récuperer les signaux d'entrées."""
        for inputs_group in self._inputs_groups.values():
            for input_ in inputs_group.inputs:
                yield input_

    @property
    def input_len (self) -> int :
        """Permet de récuperer le nombre d'entrées."""
        return sum([inputs_group.count for inputs_group in self._inputs_groups.values()])

    @property
    def inputs_groups(self) -> dict[str, InputConnectionGroupModel]:
        return self._inputs_groups

    @property
    def outputs (self) -> Iterable[OutputConnectionModel] :
        """Permet de récuperer les signaux de sorties."""
        for outputs_group in self._outputs_groups.values():
            for output in outputs_group.outputs:
                yield output

    @property
    def output_len (self) -> int :
        """Permet de récuperer le nombre de sorties."""
        return sum([outputs_group.count for outputs_group in self._outputs_groups.values()])

    @property
    def outputs_groups(self) -> dict[str, OutputConnectionGroupModel]:
        return self._outputs_groups

    @AbstractGraphicalElement.text.setter
    def text (self, text_ : str) -> None :
        AbstractGraphicalElement.text = text_
    
    @property
    def minimum_size (self) -> QSizeF :
        """Permet de récuperer la taille minimum de la boxes."""
        return QSizeF(AbstractBoxModel.MinimunBoxWidth, AbstractBoxModel.MinimunBoxHeight) if self.rotation() % 180 == 0 else QSizeF(AbstractBoxModel.MinimunBoxHeight, AbstractBoxModel.MinimunBoxWidth)

    @property
    def signals (self) -> Iterable[SignalLinkModel] :
        """Permet de récueperer la liste des liens connectés à la box."""

        _buff_already_passed = []

        for input_ in self.inputs :
            for signal in input_._links : 
                if signal in _buff_already_passed :
                    continue

                yield signal
                _buff_already_passed.append(signal)

        for output in self.outputs :
            for signal in output._links : 
                if signal in _buff_already_passed :
                    continue

                yield signal
                _buff_already_passed.append(signal)

    @property
    def size(self) -> QSizeF:
        return self.box_body.size

    @size.setter
    def size(self, new_size: QSizeF) -> None:
        self.setSize(new_size)

    # -------------
    # Methods
    # -------------
    def boundingRect(self) -> QRectF:
        if hasattr(self, "_bounding_rect"):
            return self._bounding_rect
        else:
            return super().boundingRect()

    def setIcon(self, icon):
        self.box_details.setIcon(icon)

    def setRotation (self, angle: float) -> None :
        angle = ((angle//90)%4)*90
        super().setRotation(angle)

    def setSize(self, new_size: QSizeF):
        # Handle Width
        if new_size.width() < AbstractBoxModel.MinimunBoxWidth:
            new_size.setWidth(AbstractBoxModel.MinimunBoxWidth)

        # Handle Height
        if new_size.height() < AbstractBoxModel.MinimunBoxHeight:
            new_size.setHeight(AbstractBoxModel.MinimunBoxHeight)

        if self.box_body.size != new_size:

            if self._box_type == BoxType.Math:
                self.box_body.size = new_size

            elif self._box_type == BoxType.Signal:
                self.box_body.points = [
                    QPointF(InputConnectionModel.ItemWidth, 0), 
                    QPointF(InputConnectionModel.ItemWidth, new_size.height()), 
                    QPointF(InputConnectionModel.ItemWidth + new_size.width(), new_size.height()/2)
                ]

            elif self._box_type == BoxType.Source:
                self.box_body.center = QPointF(new_size.width()/2, new_size.height()/2) + QPointF(InputConnectionModel.ItemWidth, 0)
                self.box_body.radius = min(AbstractBoxModel.MinimunBoxWidth, new_size.height()) / 2

            else:
                raise(f"AbstractBoxModel.__init__(): Box Type is not supported: {self._box_type}")

            self.update_ui()

    def setValue(self, value: Any):
        self.box_details.set_value(value)

    def add_inputs_groups(self, inputs_group: dict[str, dict], infinite_groups_reset=False):
        
        should_update_ui = False

        for name, group_data in inputs_group.items() :
            if name in self._inputs_groups:
                LogManager().error(f"{self._name}: Try to add a duplicated input group {name}")
            else:
                should_update_ui = True 
                
                # Get isInfinite parameter
                is_infinite = group_data["isInfinite"]

                # Get count parameter
                count = group_data["minimumCount"] if is_infinite and infinite_groups_reset else group_data["count"]

                # Add inputs group
                self._inputs_groups[name] = InputConnectionGroupModel(name, is_infinite, group_data["type"], 
                    group_data["minimumCount"], group_data["maximumCount"], count, parent=self)

        # Update UI
        if should_update_ui:
            self.update_ui()

    def add_inputs_group(self, name: str, group_data: dict):
        if name in self._inputs_groups:
            LogManager().error(f"{self._name}: Try to add a duplicated input group {name}")
        else:            
            # Add inputs group
            self._inputs_groups[name] = InputConnectionGroupModel(name, group_data["isInfinite"], group_data["type"], 
                group_data["minimumCount"], group_data["maximumCount"], group_data["count"], parent=self)

            self.update_ui()

    def append_input (self, group_name : str) -> InputConnectionModel :
        """Permet d'insérer une entrée dans la box."""
        if group_name in self._inputs_groups:
            # Get input group
            input_group = self._inputs_groups[group_name]

            # Append new input
            input_ = input_group.append_input()

            # Update UI
            self._should_recompute_inputs = True
            self.update_ui()

            self.input_count_changed.emit()

            return input_
        else:
            return None

    def insert_input (self, input_index : int, group_name : str) -> InputConnectionModel :
        """Permet d'insérer une entrée dans la box."""

        if group_name in self._inputs_groups:
            # Get input group
            input_group = self._inputs_groups[group_name]

            # Append new input
            input_ = input_group.insert_input(input_index)

            # Update UI
            self._should_recompute_inputs = True
            self.update_ui()

            self.input_count_changed.emit()

            return input_
        else:
            return None

    def get_inputs_group(self, group_name: Union[int, str]) -> InputConnectionGroupModel:
        if type(group_name) == int:
            group_name = list(self._inputs_groups)[group_name]

        if group_name in self._inputs_groups:
            return self._inputs_groups[group_name]

    def remove_input (self, input_group_name: str, input_ : InputConnectionModel) -> None :
        """Permet de suppprimer une entrée de la box."""

        # Remove all existing links
        for i in range(len(input_._links)) :
            self.diagram_model.remove_element(input_._links[i])

        # Remove input
        self._inputs_groups[input_group_name].remove_input(input_)
        input_.deleteLater()

        # Recompute box model
        self._should_recompute_inputs = True
        self.update_ui()

        self.input_count_changed.emit()

    def clear_inputs(self, group_name: str = None):
        # Clear all infinite input group
        if group_name is None:
            for inputs_group in self._inputs_groups:
                inputs_group.clear_inputs()
                
        # Clear by input name
        else:
            self._inputs_groups[group_name].clear_input()

    def add_outputs_groups(self, outputs_groups: dict[str, dict], infinite_groups_reset=False):
        
        should_update_ui = False

        for name, group_data in outputs_groups.items() :
            if name in self._outputs_groups:
                LogManager().error(f"{self._name}: Try to add a duplicated output group {name}")
            else:
                should_update_ui = True 
                
                # Get isInfinite parameter
                is_infinite = group_data["isInfinite"]

                # Get count parameter
                count = group_data["minimumCount"] if is_infinite and infinite_groups_reset else group_data["count"]

                # Add outputs group
                self._outputs_groups[name] = OutputConnectionGroupModel(name, is_infinite, group_data["type"], 
                    group_data["minimumCount"], group_data["maximumCount"], count, parent=self)

        # Update UI
        if should_update_ui:
            self.update_ui()

    def add_outputs_group(self, name: str, group_data: dict):
        if name in self._outputs_groups:
            LogManager().error(f"{self._name}: Try to add a duplicated output group {name}")
        else:
            # Calc x-Axis position
            x_axis_position = self.box_body.boundingRect().width() + AbstractConnectionModel.ItemWidth

            # Calculate y-axis group position
            y_group_position = sum([outputs_group.boundingRect().height() for outputs_group in self._outputs_groups.values()])
            
            # Add outputs group
            self._outputs_groups[name] = OutputConnectionGroupModel(name, group_data["isInfinite"], group_data["type"], 
                group_data["minimumCount"], group_data["maximumCount"], group_data["count"], group_position=QPointF(x_axis_position, y_group_position), parent=self)

            # Update UI
            self.update_ui()

    def append_output (self, group_name : str) -> OutputConnectionModel :
        """Permet d'insérer une entrée dans la box."""
        if group_name in self._outputs_groups:
            # Get output group
            output_group = self._outputs_groups[group_name]

            # Append new output
            output_ = output_group.append_output()

            # Update UI
            self._should_recompute_outputs = True
            self.update_ui()

            self.output_count_changed.emit()

            return output_
        else:
            return None

    def insert_output (self, output_index : int, group_name : str) -> OutputConnectionModel :
        """Permet d'insérer une entrée dans la box."""

        if group_name in self._outputs_groups:
            # Get output group
            output_group = self._outputs_groups[group_name]

            # Append new output
            output_ = output_group.insert_output(output_index)

            # Update UI
            self._should_recompute_outputs = True
            self.update_ui()

            self.output_count_changed.emit()

            return output_
        else:
            return None

    def get_outputs_group(self, group_name: Union[int, str]) -> OutputConnectionGroupModel:
        if type(group_name) == int:
            group_name = list(self._outputs_groups)[group_name]

        if group_name in self._outputs_groups:
            return self._outputs_groups[group_name]

    def remove_output (self, output_group_name: str, output_ : OutputConnectionModel) -> None :
        """Permet de suppprimer une entrée de la box."""

        # Remove all existing links
        for i in range(len(output_._links)) :
            self.diagram_model.remove_element(output_._links[i])

        # Remove output
        self._outputs_groups[output_group_name].remove_output(output_)
        output_.deleteLater()

        # Recompute box model
        self._should_recompute_outputs = True
        self.update_ui()

        self.output_count_changed.emit()

    def clear_outputs(self, group_name: str = None):
        # Clear all infinite output group
        if group_name is None:
            for outputs_group in self._outputs_groups:
                outputs_group.clear_outputs()
                
        # Clear by output name
        else:
            self._outputs_groups[group_name].clear_output()

    def update_box_size(self):
        # Calulate New Box Height
        new_box_height = max(self.input_len, self.output_len) * (AbstractConnectionModel.ItemHeight + 10)
        if new_box_height < AbstractBoxModel.MinimunBoxHeight:
            new_box_height = AbstractBoxModel.MinimunBoxHeight

        # Calculate new size
        self.setSize(QSizeF(self.box_body._size.width(), new_box_height))
        self._should_recompute_inputs = True
        self._should_recompute_outputs = True
        self._should_recompute_box_details = True
        
    def recompute_inputs_positions (self) -> None :
        """Permet de recalculer les positions des entrées."""

        # Is should not recompute → don't do anything
        if not(self._should_recompute_inputs) or len(self._inputs_groups) == 0:
            return
        self._should_recompute_inputs = False

        # Calculate total height
        biggest_io_groups = self._inputs_groups if self.input_len > self.output_len else self._outputs_groups
        total_inputs_groups_height = sum([io_group.boundingRect().height() for io_group in biggest_io_groups.values()])

        # Calculate total margin
        total_margin = self.box_body.boundingRect().height() - total_inputs_groups_height

        # Calculate margin between inputs groups
        height_margin = total_margin / (len(self._inputs_groups) + 1)

        # Init y_axis position
        y_axis_position = height_margin - 5
        
        # Recompute all inputs groups position
        for inputs_group in self._inputs_groups.values():
            # Update inputs group position
            inputs_group.setPos(QPointF(0, y_axis_position))

            # Calculate next inputs group position
            y_axis_position += inputs_group.boundingRect().height() + height_margin - 5

    def recompute_outputs_positions (self) -> None :
        """Permet de recalculer les positions des entrées."""

        # Is should not recompute → don't do anything
        if not(self._should_recompute_outputs) or len(self._outputs_groups) == 0:
            return
        self._should_recompute_outputs = False

        # Calculate total height
        biggest_io_groups = self._outputs_groups if self.input_len > self.output_len else self._outputs_groups
        total_outputs_groups_height = sum([io_group.boundingRect().height() for io_group in biggest_io_groups.values()])

        # Calculate total margin
        total_margin = self.box_body.boundingRect().height() - total_outputs_groups_height

        # Calculate margin between outputs groups
        height_margin = total_margin / (len(self._outputs_groups) + 1)

        # Init y_axis position
        y_axis_position = height_margin - 5

        # Calc x-Axis position
        x_axis_position = self.box_body.boundingRect().width() + AbstractConnectionModel.ItemWidth

        # Recompute all outputs groups position
        for outputs_group in self._outputs_groups.values():
            # Update outputs group position
            outputs_group.setPos(QPointF(x_axis_position, y_axis_position))

            # Calculate next outputs group position
            y_axis_position += outputs_group.boundingRect().height() + height_margin - 5

    def recompute_box_details(self):
        if not self._should_recompute_box_details:
            return 
        self._should_recompute_box_details = False

        # Recompute box name position
        self.name_label.setPos(self.box_body.boundingRect().bottomLeft() + self.box_body.pos())
        

        # Update box detail bounding rect
        self.box_details.set_bounding_rect(self.box_body.boundingRect().size())

    def update_ui(self):
        # Update box body size
        self.update_box_size()

        # Recompute intputs position
        self.recompute_inputs_positions()
        
        # Recompute outputs position
        self.recompute_outputs_positions()
        
        # Recompute box details
        self.recompute_box_details()

        # Recalculate bounding box
        old_bounding_rect = self.boundingRect()
        bounding_rect = self.box_body.boundingRect().translated(self.box_body.pos())
        for ios_group in [*self._inputs_groups.values(), *self._outputs_groups.values()]:
            bounding_rect = bounding_rect.united(ios_group.connectors_bounding_rect().translated(ios_group.pos()))

        self._bounding_rect = bounding_rect
        
        # Recompute intputs position
        self.update(self._bounding_rect.united(old_bounding_rect).united(self.name_label.boundingRect()))
        

    def mouseMoveEvent (self, event : QGraphicsSceneMouseEvent) -> None :
        return super().mouseMoveEvent(event)

    def mousePressEvent (self, event : QGraphicsSceneMouseEvent) -> None :
        return super().mousePressEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None :
        return super().mouseReleaseEvent(event)

    def mouseDoubleClickEvent(self, event: 'QGraphicsSceneMouseEvent') -> None:
        return super().mouseDoubleClickEvent(event)
    
    def itemChange (self, change: QGraphicsItem.GraphicsItemChange, value: Any) -> Any :
        if change == QGraphicsItem.ItemSelectedChange :
            self.selectionChanged(EventSignalData(self, value))

        # If Item is deleted
        elif change == QGraphicsItem.GraphicsItemChange.ItemSceneChange and value is None:
            if self.isUnderMouse():
                # Reinitialise Cursor 
                while QApplication.overrideCursor() is not None:
                    QApplication.restoreOverrideCursor()

        return super().itemChange(change, value)

    def selectionChanged (self, param : EventSignalData) -> None :
        # If one item is selected
        if param.value == 1 and param.sender == self:
            self.setZValue(2)

            # Update Pen
            self.box_body.line_style = Qt.PenStyle.DashLine
        else:
            self.setZValue(1)

            # Update Pen
            self.box_body.line_style = Qt.PenStyle.SolidLine

    def boxBodyHoverEnterEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        QApplication.setOverrideCursor(Qt.SizeAllCursor)
        return super().hoverEnterEvent(event)

    def boxBodyHoverLeaveEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        while QApplication.overrideCursor() is not None:
            QApplication.restoreOverrideCursor()
        return super().hoverLeaveEvent(event)


#-----------------------------------
# Imports
#-----------------------------------
# PyQt Imports
from PyQt5.QtWidgets import QWidget, QFormLayout, QLayout, QGridLayout
from PyQt5.QtCore import Qt

# Project Imports
from muphyn.packages.interface.graphical_actions.actions_holder import ActionsHolder
from muphyn.packages.interface.graphical_actions.diagram_change_element_params_action import DiagramChangeElementParamsAction
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement

#-----------------------------------
# Class
#-----------------------------------

class AbstractPropertiesEditor (QWidget) :
    """Est la classe abstraite communes aux éléments permettant de modifier un model graphique."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, model : AbstractGraphicalElement) : 

        QWidget.__init__(self, None)
        self._model = model

        self.setLayout(self.create_layout())

        self.init_ui()
        self.translate_ui()

    # -------------
    # Properties
    # -------------

    @property
    def model (self) -> AbstractGraphicalElement :
        """Permet de récuperer le modèle édité par la page de propriétés."""
        return self._model

    @property
    def actions_holder (self) -> ActionsHolder :
        """Permet de récuperer le conteneur des actions."""
        return self._actions_holder

    @actions_holder.setter
    def actions_holder (self, actions_holder_ : ActionsHolder) -> None :
        """Permet de modifier le conteneur des actions."""
        self._actions_holder = actions_holder_


    def actions_generator (self, old_value, new_value, param_name) -> None :
        """Est la méthode appelée pour générer une """

        last_action = self.actions_holder.last_action

        if not(last_action is None) :
            if isinstance(last_action, DiagramChangeElementParamsAction) :
                if last_action.param_name == param_name : 
                    if type(old_value) != bool and type(new_value) != bool:

                        last_action.new_param_value = new_value
                        return

        action = DiagramChangeElementParamsAction(self.model, param_name, new_value, old_value)
        self.actions_holder.append(action)
        
    # -------------
    # Methods
    # -------------

    def init_ui (self) -> None :
        """Permet de créer les éléments graphiques de la page de propriétés."""
        raise Exception('AbstractPropertiesElement init_ui is an abstract method and should be overridden.')

    def translate_ui (self) -> None : 
        """Permet de traduire les éléments graphiques de la page de propriétés."""
        raise Exception('AbstractPropertiesElement translate_ui is an abstract method and should be overridden.')

    def create_layout (self) -> QLayout : 
        """Permet de créer le layout pour l'affichage actuel."""
        grid_layout = QGridLayout()
        grid_layout.setColumnMinimumWidth(0, 60)
        # grid_layout.setSizeConstraint(QLayout.SizeConstraint.SetFixedSize)

        return grid_layout

    def unload (self) -> None :
        """Est la méthode appelée lorsque le page de propriété est supprimée de l'écran.""" 


#-----------------------------------
# Imports
#-----------------------------------

from typing import List
from numpy import sin

#-----------------------------------
# Functions
#-----------------------------------

def _init_sine_box (box, simulation_params) -> None :

    if not 'amplitude' in box :
        box['amplitude'] = 1.0

    if not 'pulsation' in box :
        box['pulsation'] = 1.0
    
    if not 'phase' in box :
        box['phase'] = 0.0

def _function_sine_box (box, event_) -> List : 
    
    v = box['amplitude'] * sin((event_.timing * box['pulsation']) + box['phase'] )
    
    events : List = []

    for output in box.outputs :
        events.append(box.construct_signal_event(output, v))

    return events


#-----------------------------------
# Imports
#-----------------------------------

from PyQt5.QtWidgets import QLabel, QTabWidget
from muphyn.packages.interface.editors.abstract_editor import AbstractEditor
from muphyn.packages.interface.editors.code_editor import CodeEditor
from muphyn.packages.interface.editors.multi_physics_diagram_editor import MultiPhysicsDiagramEditor
from muphyn.packages.interface.graphical_actions.actions_holder import ActionsHolder
from muphyn.packages.interface.models.editable_models.abstract_code_model import AbstractCodeModel
from muphyn.packages.interface.models.editable_models.abstract_diagram_model import AbstractDiagramModel
from muphyn.packages.interface.models.editable_models.abstract_editable_model import AbstractEditableModel
from muphyn.packages.interface.editors.diagram_editor import DiagramEditor

#-----------------------------------
# Functions
#-----------------------------------

def factory_editors (tab_holder : QTabWidget, editable_model : AbstractEditableModel) -> AbstractEditor :
    
    if hasattr(editable_model, 'editor_type') :

        if editable_model.editor_type == 'code-editor' :
            return CodeEditor(tab_holder, editable_model, ActionsHolder())
        
        elif editable_model.editor_type == 'diagram-editor' :
            return DiagramEditor(tab_holder, editable_model, ActionsHolder())

        elif editable_model.editor_type == 'multiphysics-editor' :
            return MultiPhysicsDiagramEditor(tab_holder, editable_model, ActionsHolder())

    lbl : QLabel = QLabel(tab_holder)
    lbl.setText('No editor')
    return lbl


from os import system
import time


"""

Test to evaluate if sized queue works.




Au final, cette queue n'était pas utile et a été retirée.

"""
"""
sized_queue : SizedQueue[float] = SizedQueue(10)

_ = system('cls')
for i in range(30) :

    size = len(sized_queue)
    sized_queue.push(i)
    size_2 = len(sized_queue)

    _ = system('cls')
    print('adding', i, 'to the queue', size, 'then and', size_2, 'now')
    print('')
    print('queue : ')
    for i in sized_queue :
        print(i)

    time.sleep(0.5)
"""



from PyQt5.QtCore import QMimeData, QRegExp, QSortFilterProxyModel, Qt
from PyQt5.QtWidgets import QTreeView, QAbstractItemView
from PyQt5.QtGui import QStandardItemModel, QMouseEvent, QDrag

from muphyn.packages.interface.box_library.library_element import LibraryElement
from muphyn.packages.interface.box_library.box_library_element import BoxLibraryElement, BoxLibraryElementRole
from muphyn.packages.interface.models.graphical_models.abstract_box_model import AbstractBoxModel

class LibraryTreeView(QTreeView):

    MULTIPLE_BOX_IMPORT_OFFSET_X = 0
    MULTIPLE_BOX_IMPORT_OFFSET_Y = AbstractBoxModel.MinimunBoxHeight + 20

    def __init__(self, parent = None) -> None:
        super().__init__(parent)

        # General Parameters
        self.setHeaderHidden(True)
        self.setAnimated(True)

        # Enable Drag
        self.setDragEnabled(True)
        self.setDragDropMode(QAbstractItemView.DragDropMode.DragOnly)

        # Set Selection Mode
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)

    def build_model(self, boxes_data):
        # Init Model
        tree_model: QStandardItemModel = QStandardItemModel()
        root_node = tree_model.invisibleRootItem()

        # Build Model
        model_dict = {}
        for box_data in boxes_data:
            # Split Box Library & Sub Libraries if there is a Sub Library
            library_hierarchy = box_data.box_library.split(".")
            
            # Create Component Item
            component_item = BoxLibraryElement(box_data)
            component_item.setData(box_data, BoxLibraryElementRole.BOX_DATA)

            if len(library_hierarchy) == 2:
                # Extract Library & Sub Library
                library, sub_library = library_hierarchy

                if library in model_dict.keys():
                    # Get Library Element
                    library_item: LibraryElement = model_dict[library]["library_item"]

                    # Get Sub Libraries Dictionnary
                    sub_libraries_dict = model_dict[library]["sub_libraries"]

                    if sub_library in sub_libraries_dict.keys():
                        # Get Sub Library Item
                        sub_library_item = sub_libraries_dict[sub_library]["sub_library_item"]

                        # Append Component
                        sub_library_item.appendRow(component_item)
                    else:
                        # Create Sub Library Item
                        sub_library_item = LibraryElement(sub_library)

                        sub_libraries_dict[sub_library] = {
                            "sub_library_item": sub_library_item
                        }

                        # Append Component
                        sub_library_item.appendRow(component_item)

                        # Append Sub Library
                        library_item.appendRow(sub_library_item)
                    
                else:
                    # Create Library Item
                    library_item = LibraryElement(library)

                    # Create Sub Library Item
                    sub_library_item = LibraryElement(sub_library)

                    model_dict[library] = {
                            "library_item": library_item, 
                            "sub_libraries": {
                                    sub_library: {
                                        "sub_library_item": sub_library_item
                                        }
                                }
                        }

                    # Append Component
                    sub_library_item.appendRow(component_item)

                    # Append Sub Library
                    library_item.appendRow(sub_library_item)

                    # Append Library to Root
                    root_node.appendRow(library_item)
        
        self.source_model = tree_model

        # Init Filter Model
        self.proxy_model =  QSortFilterProxyModel()
        self.proxy_model.setSourceModel(self.source_model)
        self.proxy_model.setRecursiveFilteringEnabled(True)

        self.setModel(self.proxy_model)

    def filter_library_list(self, searched_string: str):
        regex = QRegExp(searched_string, Qt.CaseSensitivity.CaseInsensitive)
        self.proxy_model.setFilterRegExp(regex)
        if searched_string != "":
            self.expandAll()
        else:
            self.collapseAll()
            for selected_index in self.selectedIndexes():
                self.expandRecursively(selected_index)
                expanded_item_parent = selected_index.parent()
                while expanded_item_parent.row() != -1:
                    self.expand(expanded_item_parent)
                    expanded_item_parent = expanded_item_parent.parent()


    def mouseMoveEvent (self, event : QMouseEvent) -> None :
        if self.selectionModel() is not None:
            for box_index, selected_index in enumerate(self.selectedIndexes()):
                # Get Selected Item
                selected_item_data = self.model().itemData(selected_index)

                # If Box Data has been saved
                if BoxLibraryElementRole.BOX_DATA in selected_item_data:
                    box_data = selected_item_data[BoxLibraryElementRole.BOX_DATA]

                    offset_x = LibraryTreeView.MULTIPLE_BOX_IMPORT_OFFSET_X * box_index
                    offset_y = LibraryTreeView.MULTIPLE_BOX_IMPORT_OFFSET_Y * box_index
                    
                    mimeData : QMimeData = QMimeData()
                    mimeData.setData('action', bytearray('new box'.encode()))
                    mimeData.setData('box', bytearray(box_data._box_name.encode()))
                    mimeData.setData('library', bytearray(box_data._box_library.encode()))
                    mimeData.setData('box_data', bytearray(str(id(box_data)).encode()))
                    mimeData.setData('offset', bytearray(str(f"{offset_x};{offset_y}").encode()))

                    drag : QDrag = QDrag(self)
                    drag.setMimeData(mimeData)
                    da : Qt.DropAction = drag.exec(Qt.CopyAction)


#-----------------------------------
# Imports
#-----------------------------------

from typing import Any

from PyQt5.QtWidgets import QGraphicsItem

from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.interface.models.links_model.abstract_link_model import AbstractLinkModel
from muphyn.packages.interface.models.links_model.link_type import LinkType
from muphyn.packages.interface.models.signals_model.abstract_signal_model import AbstractSignalModel

#-----------------------------------
# Class
#-----------------------------------

class SignalLinkModel (AbstractSignalModel, AbstractLinkModel) :
    """Est le modèle représentant les liens typés entre les boxes."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, data_type : DataType, input : Any, output : Any, link_type : LinkType = LinkType.SQUARE, 
                  text : str = '') :
                  
        AbstractSignalModel.__init__(self, data_type)
        AbstractLinkModel.__init__(self, input, output, link_type, text)

        self.setFlag(QGraphicsItem.ItemIsSelectable)

        self.b = False
        self.setZValue(0)

    # -------------
    # Properties
    # -------------

    @property
    def is_connected_to_input (self) -> bool :
        """Permet de savoir si l'élément actuel est connecté à une entrée (ou est un entrée)."""
        return False


from typing import Union

from PyQt5.QtCore import QPointF, QRectF, QSizeF, pyqtSignal, Qt
from PyQt5.QtGui import QBrush, QColor, QGradient, QPainter, QPen, QTransform
from PyQt5.QtWidgets import QGraphicsObject, QStyleOptionGraphicsItem

class AbstractShape(QGraphicsObject):
    # Drawing Area Signals
    position_changed = pyqtSignal()
    size_changed = pyqtSignal()

    # Shape Border Signals
    border_color_changed = pyqtSignal()
    border_width_changed = pyqtSignal()
    line_style_changed = pyqtSignal()

    # Filling Shape Signals
    background_color_changed = pyqtSignal()

    def __init__(self, origin: QPointF, size: QSizeF, color: QColor = Qt.GlobalColor.transparent, border_width: int = 1, 
                    line_style: Qt.PenStyle = Qt.PenStyle.SolidLine, background_color: Union[QColor, QGradient] = Qt.GlobalColor.transparent, 
                    pen_cap_style: Qt.PenCapStyle = Qt.PenCapStyle.RoundCap, pen_join_style: Qt.PenJoinStyle = Qt.PenJoinStyle.RoundJoin, parent = None):
        
        # Size
        self._size = size
        
        # Pen Color
        if color is None:
            color = Qt.GlobalColor.transparent
        self._pen = QPen(color)

        # Pen width
        if border_width is None:
            border_width = 1
        self._pen.setWidth(border_width)

        # Pen line style
        if line_style is None:
            line_style = Qt.PenStyle.SolidLine
        self._pen.setStyle(line_style)

        # Pen cap/join style
        self._pen.setCapStyle(pen_cap_style)
        self._pen.setJoinStyle(pen_join_style)

        # Determine type of Brush based on type of Color
        brush_style = Qt.BrushStyle.SolidPattern 
        if isinstance(background_color, QGradient):
            if background_color.type() == QGradient.Type.LinearGradient:
                brush_style = Qt.BrushStyle.LinearGradientPattern
            elif background_color.type() == QGradient.Type.RadialGradient:
                brush_style = Qt.BrushStyle.RadialGradientPattern
            elif background_color.type() == QGradient.Type.ConicalGradient:
                brush_style = Qt.BrushStyle.ConicalGradientPattern
            else:
                brush_style = Qt.BrushStyle.NoBrush
        # Create brush
        self._brush = QBrush(background_color, brush_style)

        # parent
        super().__init__(parent)
        
        # Set pos
        self.setPos(origin)

        # Set All transformation origin to center of bounding box
        if size is not None:
            self.setTransformOriginPoint(QRectF(QPointF(), size).center())
    
    # -------------
    # Properties
    # -------------
    @property
    def position(self) -> QPointF:
        return self.pos()

    @position.setter
    def position(self, new_position) -> None:
        if new_position != self.pos():
            # Update position
            self.setPos(new_position)

            # Emit modification signals
            self.position_changed.emit()

    @property
    def size(self) -> QPointF:
        return self._size
            
    @property
    def border_color(self) -> QColor:
        return self._pen.color()

    @border_color.setter
    def border_color(self, new_color: QColor) -> None:
        if self._pen.color() != new_color:
            self._pen.setColor(new_color)

            self.border_color_changed.emit()

    @property
    def border_width(self) -> int:
        return self._pen.width()

    @border_width.setter
    def border_width(self, new_border_width: int) -> None:
        if self._pen.width() != new_border_width:
            self._pen.setWidth(new_border_width)

            self.border_width_changed.emit()

    @property
    def line_style(self) -> Qt.PenStyle:
        return self._pen.style()

    @line_style.setter
    def line_style(self, new_line_style: Qt.PenStyle) -> None:
        if self._pen.style() != new_line_style:
            self._pen.setStyle(new_line_style)

            self.line_style_changed.emit()

    @property
    def background_color(self) -> QColor:
        return self._brush.color()

    @background_color.setter
    def background_color(self, new_fill_color: Union[QColor, QGradient]) -> None:
        
        if isinstance(new_fill_color, Qt.GlobalColor):
            new_fill_color = QColor(new_fill_color)
            
        if isinstance(new_fill_color, QGradient):
            if new_fill_color != self._brush.gradient():
                # Select Brush Style
                brush_style = Qt.BrushStyle.NoBrush
                if new_fill_color.type() == QGradient.Type.LinearGradient:
                    brush_style = Qt.BrushStyle.LinearGradientPattern
                elif new_fill_color.type() == QGradient.Type.RadialGradient:
                    brush_style = Qt.BrushStyle.RadialGradientPattern
                elif new_fill_color.type() == QGradient.Type.ConicalGradient:
                    brush_style = Qt.BrushStyle.ConicalGradientPattern

                # Create new brush
                self._brush = QBrush(new_fill_color, brush_style)

                self.background_color_changed.emit()

        elif isinstance(new_fill_color, QColor):
            if new_fill_color != self._brush.color():
                self._brush.setColor(new_fill_color)
                self._brush.setStyle(Qt.BrushStyle.SolidPattern)

                self.background_color_changed.emit()

    @property
    def pen(self) -> QPen:
        return self._pen

    @pen.setter
    def pen(self, new_pen: QPen) -> None:
        self._pen = new_pen

    @property
    def brush(self) -> QBrush:
        return self._brush

    @brush.setter
    def brush(self, new_brush: QPen) -> None:
        self._brush = new_brush

    # -------------
    # Methods
    # -------------
    def boundingRect(self) -> QRectF:
        return QRectF(QPointF(), self._size)

    def draw_shape(self, painter : QPainter, option: QStyleOptionGraphicsItem, widget) -> None:
        raise(NotImplementedError(f"{type(self).__name__}: AbstractShape.draw_shape(painter, option, widget) not reimplemented"))

    def paint (self, painter : QPainter, option: QStyleOptionGraphicsItem = QStyleOptionGraphicsItem.OptionType.SO_Frame, widget=None) -> None :
        # Save painter state
        painter.save()

        # Set Pen
        painter.setPen(self._pen)

        # Set Brush
        painter.setBrush(self._brush)

        # Draw Shape
        self.draw_shape(painter, option, widget)

        # Reset painter state
        painter.restore()



#-----------------------------------
# Imports
#-----------------------------------

from PyQt5.QtWidgets import QListWidget, QListWidgetItem
from muphyn.packages.interface.graphical_actions.abstract_graphical_action import AbstractGraphicalAction

#-----------------------------------
# Functions
#-----------------------------------

def _lst_contains_element (lst : QListWidget, el : str) -> bool :
    """Est la méthode appelée pour savoir si un élément est déjà contenu dans la liste."""
    
    for i in range(lst.__len__()) :
        item = lst.item(i)

        if item.text() == el :
            return True 

    return False

#-----------------------------------
# Class
#-----------------------------------

class LibraryDialogAddAction (AbstractGraphicalAction) :
    """Est l'action défaisable pour ajouter des éléments dans la liste."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, lst : QListWidget, element : str) :
        self._lst = lst
        self._element : str = element

    # -------------
    # Methods
    # -------------

    def do (self) :
        
        if not _lst_contains_element(self._lst, self._element) :
            QListWidgetItem(self._element, self._lst)

    def undo (self) :
        
        for i in range(self._lst.__len__()) :
            item = self._lst.item(i)

            if item.text() == self._element :
                self._lst.takeItem(i)
                return


#-----------------------------------
# Imports
#-----------------------------------
# General Imports
from typing import Generic, Iterable, Type, TypeVar, List, Any

# PyQt5 Imports
from PyQt5 import QtGui
from PyQt5.QtCore import QPointF, QSizeF, pyqtSignal
from PyQt5.QtWidgets import QGraphicsItem, QStyleOptionGraphicsItem

# Project Imports
from muphyn.packages.interface.models.graphical_models.abstract_moveable_graphical_element import AbstractMoveableGraphicalElement
from muphyn.packages.interface.models.links_model.abstract_link_model import AbstractLinkModel

#-----------------------------------
# Generic Type
#-----------------------------------

T = TypeVar('T', bound = AbstractLinkModel)

#-----------------------------------
# Class
#-----------------------------------


class AbstractNodeModel (Generic[T], AbstractMoveableGraphicalElement) :
    """Est la classe abstraite commune aux noeuds capables d'être affiché dans l'interface."""
    
    
    # -------------
    # Signals
    # -------------
    links_count_changed = pyqtSignal()

    # -------------
    # Constructors
    # -------------

    def __init__ (self, name : str, position : QPointF, size : QSizeF, links : Iterable[Type[T]],
                  text : str = '', parent : QGraphicsItem = None) :

        AbstractMoveableGraphicalElement.__init__(self, name, position, size, 0, text, parent)

        self._links : List[Type[T]] = []

        self.connector_center = QPointF(0, 0)
        
        self.setFlag(QGraphicsItem.ItemIsMovable, False)

        for link in links :
            self._links.append(link)
        self.links_count_changed.emit()

    # -------------
    # Properties
    # -------------

    @property 
    def links (self) -> Iterable[Type[T]] :
        """Permet de récuperer la liste des liens connectés au noeud."""

        for link in self._links :
            yield link

    # -------------
    # Methods
    # -------------
 
    def __len__ (self) -> int :
        """Permet de récuperer le nombre de liens connectés au noeud."""
        return self._links.__len__()

    def setRotation (self, angle: float) -> None :
        ...

    def add_link (self, link : Any) -> None :
        """Permet d'ajouter un lien à la position données."""

        if link is None : 
            return

        self._links.append(link)
        self.links_count_changed.emit()

    def insert_link (self, index : int, link : Any) -> None :
        """Permet d'insérer un lien à la position données."""

        if link is None : 
            return

        if index > len(self._links) :
            return

        self._links.insert(index, link)
        self.links_count_changed.emit()

    def remove_link (self, link : Type[T]) -> None :
        """Permet de supprimer un lien."""

        if link is None : 
            return

        self._links.remove(link)
        self.links_count_changed.emit()
    
    def paint (self, painter: QtGui.QPainter, option: QStyleOptionGraphicsItem, widget) -> None:
        ...


#-----------------------------------
# Imports
#-----------------------------------

from typing import List, Dict

from .plci_core_box import Box
from .plci_core_signal import Signal

#-----------------------------------
# Class
#-----------------------------------

class Diagram :

    # -------------
    # Constructors
    # -------------
    def __init__ (self) :
        self._boxes : List[Box] = []
        self._signals : List[Signal] = []
        self._box_inputs : Dict[Signal, List[Box]] = {}
        self._linked_signals : Dict[Signal, List[Signal]] = {}
        self._inputs : List[int] = []
        self._outputs : List[int] = []

    # -------------
    # Properties
    # -------------

    @property
    def boxes (self) -> List[Box] :
        """Permet de récuperer les boxes dans le diagramme."""
        return self._boxes

    @property
    def signals (self) -> List[Signal] :
        """Permet de récuperer les signaux dans le diagramme."""
        return self._signals

    @property 
    def box_inputs (self) -> Dict[Signal, List[Box]] :
        """Permet de récuperer la liste des sigaux et les boites qui en dépendent."""
        return self._box_inputs

    @property
    def linked_signals (self) -> Dict[Signal, List[Signal]] :
        """Permet de récuperer la liste des connexions signaux entre le signaux."""
        return self._linked_signals

    @property
    def inputs (self) -> List[int] :
        """Permet de récuperer la liste des index des signaux d'entrées."""
        return self._inputs

    @property 
    def outputs (self) -> List[int] :
        """Permet de récuperer la liste des index des signaux de sorties."""
        return self._outputs

    # -------------
    # Methods
    # -------------

    def append (self, obj) -> None :
        """Permet d'ajouter un objet au diragramme."""

        if isinstance(obj, Diagram) :
            for box in obj.boxes :
                self._boxes.append(box)

            for signal in obj.signals :
                self._signals.append(signal)

            for signal in obj.box_inputs.keys() :
                self._box_inputs[signal] = obj.box_inputs[signal]

            for signal in obj._linked_signals.keys() :
                self._linked_signals[signal] = obj._linked_signals[signal]

        elif isinstance(obj, Box) :
            self.boxes.append(obj)

        elif isinstance(obj, Signal) :
            self.signals.append(obj)


    def add_box_inputs (self, box : Box, *signals : Signal) -> None :
        """Permet de lier des signaux comme entrée à une box."""
        
        signals_to_compute = []
        
        if not box in self._boxes :
            raise Exception(box, " not in diagram.")

        for signal in signals:

            if not signal in self._signals :
                raise Exception(signal, " not in diagram.")

            signals_to_compute.append(signal)

        for signal_to_compute in signals_to_compute :

            box.append_inputs(signal_to_compute)

            if not signal_to_compute in self._box_inputs:
                self._box_inputs[signal_to_compute] = []
            
            self._box_inputs[signal_to_compute].append(box)
        

    def add_box_outputs (self, box : Box, *signals : Signal) -> None :
        """Permet de lier des signaux comme sortie à une box."""

        signals_to_compute = []

        if not box in self._boxes :
            raise Exception(box, " not in diagram.")
            return

        for signal in signals:

            if not signal in self._signals :
                raise Exception(signal, " not in diagram.")

            signals_to_compute.append(signal)
                
        for signal_to_compute in signals_to_compute:
            box.append_outputs(signal_to_compute)
        
        
    def add_linked_signals (self, signal : Signal, *signals : Signal) -> None :
        """Permet de lier des signaux entre eux pour générer des modifications de valeurs en cascade."""

        signals_to_compute : List[Signal] = []

        if not signal in self._signals :
            Exception(signal, " not in diagram.")
            return

        for current_signal in signals:

            if not current_signal in self._signals :
                raise Exception(current_signal, " not in diagram.")
                
            signals_to_compute.append(current_signal)
            
        if not signal in self._linked_signals :
            self._linked_signals[signal] = []

        for signal_to_compute in signals_to_compute :
            self._linked_signals[signal].append(signal_to_compute)

        


        



#-----------------------------------
# Imports
#-----------------------------------

from typing import Iterable, overload
from PyQt5.QtGui import QIcon
from PyQt5.QtCore import pyqtSignal, pyqtSlot
from PyQt5.QtWidgets import QTabWidget, QWidget

from muphyn.packages.interface.editors.abstract_editor import AbstractEditor
from muphyn.packages.interface.editors.diagram_editor import DiagramEditor
from muphyn.packages.core.utils.log_manager import LogManager

#-----------------------------------
# Class
#-----------------------------------

class ProjectTabsHolder (QTabWidget) :
    """Est le widget qui affiche les différents onglets des projets du programme."""
    

    # -------------
    # Signals
    # -------------
    
    elements_selected_changed = pyqtSignal(object)
    """Est le signal appelé quand l'onglet sélectionné est changé."""

    tab_close_request = pyqtSignal(object)
    """Est le signal appelée quand un des onglet reçoit une requête de fermeture."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, parent : QWidget) :
        
        QTabWidget.__init__(self, parent)

        self.setTabsClosable(True)
        self.setMovable(True)
        self.currentChanged.connect(self.current_tab_changed)
        self._current_editor = None 

    # -------------
    # Properties
    # -------------

    @property
    def current_editor (self) -> AbstractEditor : 
        """Permet de récuperer l'éditeur actuel."""
        return self._current_editor

    @current_editor.setter
    def current_editor (self, current_editor_ : AbstractEditor) -> None :
        """Permet de modifier l'éditeur actuel."""

        if not(self._current_editor is None) : 
            self._current_editor.elements_selected_changed.disconnect(self.current_editor_selected_elements_changed)

        self._current_editor = current_editor_

        if self._current_editor is None :
            self.elements_selected_changed.emit(None)
            
        else : 
            self.elements_selected_changed.emit(self._current_editor.selected_elements)
            self._current_editor.elements_selected_changed.connect(self.current_editor_selected_elements_changed) 

    # -------------
    # Methods
    # -------------
    def __iter__(self) -> Iterable[DiagramEditor]:
        return iter([self.widget(childIndex) for childIndex in range(self.count())])

    def current_tab_changed (self, index: int) -> None :
        """Est la méthode appelée lorsque l'utilisateur modifie l'onglet selectionné."""

        if self.currentWidget() == None :
            self.current_editor = None 

        elif isinstance(self.currentWidget(), AbstractEditor) :
            # Change the current editor
            self.current_editor = self.currentWidget()

            # Update in the LOG Manager the current editor
            LogManager().set_current_project_id(self.currentWidget().id)

    def current_editor_selected_elements_changed (self, elements) -> None :
        """Est la méthode appelée lorsque la sélection de l'utilisateur est modifiée."""
        self.elements_selected_changed.emit(elements)

    def addEditor (self, model : AbstractEditor) -> int :
        model.name_changed.connect(self.tab_name_changed)
        return super().addTab(model, model.tab_text)

    def removeTab (self, index : int) -> None :
        editor = self.widget(index)
        if hasattr(editor, 'name_changed') :
            editor.name_changed.disconnect(self.tab_name_changed)

        super().removeTab(index)
    
    def tab_name_changed (self, tab_ : AbstractEditor, title_ : str) -> None :
        """Est la méthode appelée lorsque le nom du modèle est modifié."""
        LogManager().debug('tab name changed')
        self.setTabText(self.indexOf(tab_), title_)


#-----------------------------------
# Imports
#-----------------------------------
from muphyn.packages.core.utils.singleton import SingletonMetaClass


class GlobalEnvVariablesManager(metaclass=SingletonMetaClass) :
    """Est la classe qui permet de construire les boxes.""" 

    # -------------
    # Constructors
    # -------------
    def __init__ (self, global_vars: dict) :
        self.global_vars: dict = global_vars


#-----------------------------------
# Imports
#-----------------------------------

# General Imports
from typing import Iterable, List

# PyQt5 Imports
from PyQt5.QtWidgets import QGraphicsScene

# Project Imports
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.interface.models.graphical_models.abstract_box_model import AbstractBoxModel
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement
from muphyn.packages.interface.models.links_model.link_type import LinkType
from muphyn.packages.interface.models.signals_model.input_connection_model import InputConnectionModel
from muphyn.packages.interface.models.signals_model.output_connection_model import OutputConnectionModel
from muphyn.packages.interface.models.signals_model.signal_link_model import SignalLinkModel
from muphyn.packages.interface.graphical_actions.actions_holder import ActionsHolder

#-----------------------------------
# Class
#-----------------------------------

class AbstractDiagramModel :
    """ESt la classe abstraite commune aux modèles des éléments graphiques capables de modifier des diagrammes."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, graphical_elements : Iterable[AbstractGraphicalElement] = []) :

        self._scene = None
        self._graphical_elements : List[AbstractGraphicalElement] = list(graphical_elements)

    # -------------
    # Properties
    # -------------

    @property
    def actions_holder (self) -> ActionsHolder :
        """Permet de récuperer le conteneur d'action."""
        if hasattr(self._scene, 'actions_holder') :
            return self._scene.actions_holder

        return None
    
    @property
    def editor_type (self) -> str :
        """Permet de récuperer le type d'éditeur à utiliser."""
        return "diagram-editor"

    @property
    def graphical_elements (self) -> Iterable[AbstractGraphicalElement] :
        """Permet de récuperer la liste des éléments graphiques affichés dans l'interface graphique."""
        return self._graphical_elements
        
    @property
    def selected_elements (self) -> Iterable[AbstractGraphicalElement] :
        """Permet de récuperer la liste des éléments graphiques sélectionnés."""

        for graphic_element in self._graphical_elements :

            if graphic_element.isSelected() :
                yield graphic_element
                
    @property
    def elements_len (self) -> int :
        """Permet de récuperer le nombre d'éléments graphiques contenus dans le modèle."""
        return self._graphical_elements.__len__()

    @property
    def box (self) -> Iterable[AbstractBoxModel] :
        """Permet de récuperer la liste des éléments graphiques qui sont des boxes."""
        
        for graphic_element in self._graphical_elements :
            if isinstance(graphic_element, AbstractBoxModel) :
                yield graphic_element

    @property
    def scene (self) -> QGraphicsScene :
        """Permet de récuperer la scène dans auquel est liée le modèle."""
        return self._scene

    @scene.setter
    def scene (self, scene_ : QGraphicsScene) -> None : 
        """Permet de modifier la scène du diagramme."""

        if not(self._scene is None) :
            for el in self._graphical_elements :
                if el in self._scene.items() :
                    self._scene.removeItem(el) 

        self._scene = scene_

        if not(self._scene is None) :
            
            for el in self._graphical_elements : 
                if not(el in self._scene.items()) :
                    self._scene.addItem(el)

    # -------------
    # Methods
    # -------------

    def __len__ (self) -> int :
        """Permet de récuperer le nombre d'éléments contenus dans le modèle de diagramme."""
        return self._graphical_elements.__len__()

    def add_element (self, graphical_element : AbstractGraphicalElement) -> None :
        """Permet d'ajouter un élément au modèle."""
        
        if graphical_element is None :
            return

        self._graphical_elements.append(graphical_element)
        graphical_element.diagram_model = self
        
        if not(self._scene is None) :
            self._scene.addItem(graphical_element)

    def remove_element (self, graphical_element : AbstractGraphicalElement) -> None : 
        """Permet de supprimer un élément du modèle."""
        
        if graphical_element is None : 
            return
        
        if not(self._scene is None) :
            self._scene.removeItem(graphical_element)
            graphical_element.deleteLater()

        graphical_element.diagram_model = None
        self._graphical_elements.remove(graphical_element)

    def link_nodes (self, input_, output, input_index : int = -1, output_index : int = -1, link_value : float = 0.5, link_type : LinkType = LinkType.SQUARE, link_text : str = '') -> SignalLinkModel :
        """Permet de relier deux noeuds."""
            
        if input_ is None :
            return None

        if output is None :
            return None

        if input_index == -1 :
            input_index = len(input_)

        if output_index == -1 :
            output_index = len(output)

        if isinstance(input_, InputConnectionModel) and isinstance(output, OutputConnectionModel) :
            
            if not(input_.data_type == output.data_type) :
                LogManager().error('input_ and output does not have the same type.')
                return

            if len(input_) > 0 : 
                LogManager().error('the input_ has already a link')
                return
            
            link : SignalLinkModel = SignalLinkModel(output.data_type, input_, output, link_type, link_text)
            link.link_value = link_value

            output.insert_link(output_index, link)
            input_.add_link(link)
            
            self.add_element(link)

            return link
        
        else :
            LogManager().error('input_ and output cannot be bound.')

    def get_element_by_graphical_index (self, graphical_index : int) -> AbstractGraphicalElement : 
        """Permet de récuperer un élément graphique grâce à son index."""

        for el in self.graphical_elements :
            if el.graphical_index == graphical_index :
                return el

        for el in self.graphical_elements :
            if isinstance(el, AbstractBoxModel) :
                for input_ in el.inputs :
                    if input_.graphical_index == graphical_index :
                        return input_

                for output in el.outputs :
                    if output.graphical_index == graphical_index :
                        return output

        return None



# General imports
from typing import Any, Union

# PyQt5 imports
from PyQt5.QtCore import QMarginsF, QPointF, QRectF, QSizeF, Qt
from PyQt5.QtWidgets import QGraphicsItem

# Project imports
from muphyn.packages.interface.shapes.basic_shape import AbstractImage, GroupedShapes, Text, VerticalLayout
from muphyn.packages.interface.utils.graphical_items import Icon
from muphyn.packages.interface.utils.interface_constants import MuphynFonts
from muphyn.packages.core.utils.log_manager import LogManager

class BoxModelInfoArea(GroupedShapes):

    # -------------
    # Constants
    # -------------
    ItemSpacing = 5

    def __init__(self, box_body_size: QSizeF, name: str, image_path: str = None, 
            value: Any = None, margins: Union[int, float, QMarginsF] = None, parent: QGraphicsItem = None) -> None:
        
        # Handle box details
        self._value = value
        self._image_path: str = image_path
        self._name: str = name

        # Init Grouped Shapes
        super().__init__(QPointF(), parent=parent)

        # Handle Margins
        self.set_margins(margins)

        # Calculate marginized bounding rect
        box_body_bounding_rect = QRectF(QPointF(0, 0), box_body_size)
        marginized_bounding_rect = box_body_bounding_rect.marginsRemoved(self.margins)

        # Add Layout
        self.layout = VerticalLayout(group_position=marginized_bounding_rect.topLeft(), parent=self)
        self.layout.setMaximumWidth(marginized_bounding_rect.size().width())
        self.layout.setMaximumHeight(marginized_bounding_rect.size().height())
        self.layout.setSpacing(BoxModelInfoArea.ItemSpacing)

        if image_path is not None:
            # Icon
            self.icon_shape: AbstractImage = Icon(image_path)

            # Add to layout
            self.layout.addItem(self.icon_shape)

        if value is not None:
            # Text Flag
            self.text_align = Qt.AlignmentFlag.AlignHCenter

            # Value Label
            self.value_label: Text = Text(
                str(value), text_max_width=self.layout.maximumWidth,
                font=MuphynFonts.BoxModelDetailsFont, alignment=self.text_align)
                
            # Add to layout
            self.layout.addItem(self.value_label, height=self.value_label.size.height())

    def set_value(self, value: str):
        if not hasattr(self, "value_label"):
                        # Text Flag
            self.text_align = Qt.AlignmentFlag.AlignHCenter

            # Value Label
            self.value_label: Text = Text(
                str(value), text_max_width=self.layout.maximumWidth,
                font=MuphynFonts.BoxModelDetailsFont, alignment=self.text_align)
                
            # Add to layout
            self.layout.addItem(self.value_label, height=self.value_label.size.height())
        else:
            self.value_label.text = str(value)

    def setIcon(self, icon):
        if not hasattr(self, "icon_shape"):
            # Icon
            self.icon_shape: AbstractImage = Icon(icon)

            # Add to layout
            self.layout.addItem(self.icon_shape)

    def set_bounding_rect(self, new_bounding_rect_size: QSizeF):
        # 
        marginized_bounding_rect_size = new_bounding_rect_size.shrunkBy(self.margins)

        # Update max height/width
        self.layout.setMaximumWidth(marginized_bounding_rect_size.width())
        self.layout.setMaximumHeight(marginized_bounding_rect_size.height())

        # Update layout child items position
        self.layout.recomputeChildrenPosition()

    def set_margins(self, new_margins: Union[int, float, QMarginsF]):
        # Handle margins format
        if new_margins is None:
            new_margins = QMarginsF(0, 0, 0, 0)
        elif type(new_margins) == float or type(new_margins) == int:
            new_margins = QMarginsF(new_margins, new_margins, new_margins, new_margins)
        elif type(new_margins) != QMarginsF:
            LogManager().error(f"BoxModelInfoArea: Wrong margins format {type(new_margins)} instead of {type(QMarginsF)}")
            return

        self.margins = new_margins


from typing import Union

from PyQt5.QtCore import QFileInfo, QPointF, QRectF, QSize, QSizeF, Qt, pyqtSignal
from PyQt5.QtGui import QColor, QGradient, QImage, QImageReader, QPainter, QTransform
from PyQt5.QtWidgets import QStyleOptionGraphicsItem
from PyQt5.QtSvg import QSvgRenderer

from muphyn.packages.interface.shapes.basic_shape.abstract_shape import AbstractShape
from muphyn.packages.core.utils.log_manager import LogManager

class AbstractImage(AbstractShape):

    # -------------
    # Signals
    # -------------
    image_path_changed = pyqtSignal() 

    def __init__(self, image_path: str, position: QPointF = QPointF(0, 0), size: QSize = None, 
            border_width: int = 0, border_color: Union[QColor, QGradient] = Qt.GlobalColor.transparent, border_line_style: Qt.PenStyle = Qt.PenStyle.SolidLine, 
            background_color: Union[QColor, QGradient] = Qt.GlobalColor.transparent, 
            aspect_ratio_mode: Qt.AspectRatioMode = Qt.AspectRatioMode.IgnoreAspectRatio, transformation_mode: Qt.TransformationMode = Qt.TransformationMode.FastTransformation, 
            parent=None):

        # Test if image file exists
        if not QFileInfo(image_path).exists():
            raise(FileNotFoundError(f"File doesn't exists:\n\t{image_path}"))

        self._image_path: str = image_path

        self._aspect_ratio_mode: Qt.AspectRatioMode = aspect_ratio_mode
        self._transformation_mode: Qt.TransformationMode = transformation_mode

        super().__init__(position, size, border_color, border_width, border_line_style, background_color, parent=parent)

        # Connect signals
        self.image_path_changed.connect(self._on_image_path_changed)

    # -------------
    # Properties
    # -------------
    @property
    def image_path(self) -> str:
        return self._image_path

    @image_path.setter
    def image_path(self, new_image_path: str) -> None:
        if self._image_path != new_image_path:
            self._image_path = new_image_path

    @property
    def aspect_ratio_mode(self) -> Qt.AspectRatioMode:
        return self._aspect_ratio_mode

    @aspect_ratio_mode.setter
    def image_path(self, new_aspect_ratio_mode: Qt.AspectRatioMode) -> None:
        if self._aspect_ratio_mode != new_aspect_ratio_mode:
            self._aspect_ratio_mode = new_aspect_ratio_mode

    @property
    def transformation_mode(self) -> Qt.TransformationMode:
        return self._transformation_mode

    @transformation_mode.setter
    def transformation_mode(self, new_transformation_mode: Qt.TransformationMode) -> None:
        if self._transformation_mode != new_transformation_mode:
            self._transformation_mode = new_transformation_mode

    @property
    def painted_area(self) -> QRectF:
        return self._painted_area

    # -------------
    # Methods
    # -------------
    def _on_image_path_changed(self):
        raise(NotImplementedError(f"{type(self).__name__}._on_image_path_changed() not overriden!"))

    def setSize(self, new_size: Union[QSize, QSizeF]):
        raise(NotImplementedError(f"{type(self).__name__}.setSize() not overriden!"))

class Image(AbstractImage):
    def __init__(self, image_path: str, position: QPointF = QPointF(0, 0), size: QSize = None, 
            border_width: int = 1, border_color: Union[QColor, QGradient] = Qt.GlobalColor.transparent, border_line_style: Qt.PenStyle = Qt.PenStyle.SolidLine, 
            background_color: Union[QColor, QGradient] = Qt.GlobalColor.transparent, 
            aspect_ratio_mode: Qt.AspectRatioMode = Qt.AspectRatioMode.IgnoreAspectRatio, transformation_mode: Qt.TransformationMode = Qt.TransformationMode.FastTransformation, 
            parent=None):

        super().__init__(image_path, position, size, 
                            border_width, border_color, border_line_style, 
                            background_color, 
                            aspect_ratio_mode, transformation_mode, parent)

        # Load Image & calc Painted Area
        self._load_image()
    

    # -------------
    # Methods
    # -------------
    def boundingRect(self) -> QRectF:
        return self._painted_area

    def _load_image(self):
         # Load Image
        image = QImage(self._image_path)

        # Save original image in memory
        self._image = image

        # Resize Image if asked
        if self._size is not None:
            image = image.scaled(
                self._size, 
                aspectRatioMode=self._aspect_ratio_mode, 
                transformMode=self._transformation_mode
            )
        else:
            self.size = image.size()
        
        # Building painted Area
        painted_area_size: QSize = image.size()
        image_to_bounding_rect_size_diff: QSize = (self._size.toSize() - painted_area_size) / 2
        painted_area_position: QPointF = QPointF(
            image_to_bounding_rect_size_diff.width(), 
            image_to_bounding_rect_size_diff.height()
        )
        
        self._painted_area = QRectF(painted_area_position, QSizeF(painted_area_size))

    def draw_shape(self, painter: QPainter, option: QStyleOptionGraphicsItem, widget) -> None:
        # Init Generator
        painter.drawImage(self._painted_area, self._image)

    def setSize(self, new_size: Union[QSize, QSizeF]):
        # Handle new size type
        new_size = QSizeF(new_size) if type(new_size) == QSize else new_size

        if new_size != self._size:
            # Update new size
            self._size = new_size

            # Resize Image
            image = self._image.scaled(
                new_size, 
                aspectRatioMode=self._aspect_ratio_mode, 
                transformMode=self._transformation_mode
            )
            
            # Building painted Area
            painted_area_size: QSize = image.size()
            image_to_bounding_rect_size_diff: QSize = (self._size.toSize() - painted_area_size) / 2
            painted_area_position: QPointF = QPointF(
                image_to_bounding_rect_size_diff.width(), 
                image_to_bounding_rect_size_diff.height()
            )
            
            self._painted_area = QRectF(painted_area_position, QSizeF(painted_area_size))

    # -------------
    # Slots
    # -------------
    def _on_image_path_changed(self):
        self._load_image()

class SVGImage(AbstractImage):
    def __init__(self, image_path: str, position: QPointF = QPointF(0, 0), size: QSize = None, 
            border_width: int = 0, border_color: Union[QColor, QGradient] = Qt.GlobalColor.transparent, border_line_style: Qt.PenStyle = Qt.PenStyle.SolidLine, 
            background_color: Union[QColor, QGradient] = Qt.GlobalColor.transparent, 
            aspect_ratio_mode: Qt.AspectRatioMode = Qt.AspectRatioMode.IgnoreAspectRatio, transformation_mode: Qt.TransformationMode = Qt.TransformationMode.FastTransformation, 
            parent = None):

        super().__init__(image_path, position, size, 
                            border_width, border_color, border_line_style, 
                            background_color, 
                            aspect_ratio_mode, transformation_mode, parent)

        # Load SVG file
        svg_renderer = QSvgRenderer()
        self._svg_renderer = svg_renderer
        self._load_svg_file()

    # -------------
    # Methods
    # -------------
    def boundingRect(self) -> QRectF:
        return self._painted_area.translated(self.pos())

    def _load_svg_file(self):
        self._svg_renderer.load(self._image_path)
        if not self._svg_renderer.isValid():
            raise(Exception("Not a valid SVG file"))

        # Convert size (QSizeF → QSize)
        if self._size is not None:
            self._size = self._size.toSize()
        else:
            self._size = self._svg_renderer.defaultSize()

        # Building painted Area
        painted_area_size: QSize = self._svg_renderer.defaultSize().scaled(
            self._size, self._aspect_ratio_mode)
        image_to_bounding_rect_size_diff: QSize = (self._size - painted_area_size) / 2
        painted_area_position: QPointF = QPointF(
            image_to_bounding_rect_size_diff.width(), 
            image_to_bounding_rect_size_diff.height()
        )
        
        self._painted_area = QRectF(painted_area_position, QSizeF(painted_area_size))

    def setWidth(self, new_width: int) -> None:
        if new_width != self._size.width():
            self.setSize(QSize(new_width, self._size.height()))

    def setHeight(self, new_height: int) -> None:
        if new_height != self._size.height():
            self.setSize(QSize(self._size.width(), new_height))

    def setSize(self, new_size: Union[QSize, QSizeF]) -> None:
        # Handle new size type
        new_size = new_size if type(new_size) == QSize else new_size.toSize()
        
        if new_size != self._size:
            # Replace value
            self._size = new_size

            # Calculate painted area size
            painted_area_size: QSize = self._svg_renderer.defaultSize().scaled(
                self._size, self._aspect_ratio_mode)
            
            # Painted area position
            image_to_bounding_rect_size_diff: QSize = (self._size - painted_area_size) / 2
            painted_area_position: QPointF = QPointF(
                image_to_bounding_rect_size_diff.width(), 
                image_to_bounding_rect_size_diff.height()
            )
            self._painted_area = QRectF(painted_area_position, QSizeF(painted_area_size))

    def draw_shape(self, painter: QPainter, option: QStyleOptionGraphicsItem, widget) -> None:
        # Init Generator
        self._svg_renderer.render(painter, self._painted_area)

    # -------------
    # Slots
    # -------------
    def _on_image_path_changed(self):
        self._load_svg_file()

def ImageShapeFactory(
        image_path: str, 
        position: QPointF = QPointF(0, 0),
        size: QSize = None,
        border_width: int = 0,
        border_color: Union[QColor, QGradient] = Qt.GlobalColor.transparent,
        border_line_style: Qt.PenStyle = Qt.PenStyle.SolidLine,
        background_color: Union[QColor, QGradient] = Qt.GlobalColor.transparent,
        aspect_ratio_mode: Qt.AspectRatioMode = Qt.AspectRatioMode.KeepAspectRatio,
        transformation_mode: Qt.TransformationMode = Qt.TransformationMode.FastTransformation,
        parent=None
    ) -> AbstractImage:

    if QFileInfo(image_path).exists():
    
        if (image_format := QImageReader.imageFormat(image_path)) in QImageReader.supportedImageFormats():
            if image_format == b"svg":
                return SVGImage(
                    image_path, position, size, border_width, border_color, border_line_style,
                    background_color, aspect_ratio_mode, transformation_mode, parent)
            else:
                return Image(
                    image_path, position, size, border_width, border_color, border_line_style,
                    background_color, aspect_ratio_mode, transformation_mode, parent)
                        
        else:
            LogManager().error(f"Error while trying to create AbstractImage object : \n\tImage file format of \"{image_path}\" is not a supported image file format")

    else:
        LogManager().error(f"Error while trying to create AbstractImage object : \n\tPath \"{image_path}\" doesn't exists")


#-----------------------------------
# Imports
#-----------------------------------

from datetime import date
from typing import Iterable
from muphyn.packages.interface.models.graphical_models.abstract_box_model import AbstractBoxModel
from muphyn.packages.interface.models.editable_models.abstract_code_model import AbstractCodeModel
from muphyn.packages.interface.models.graphical_models.box_input_model import BoxInputModel
from muphyn.packages.interface.models.graphical_models.box_output_model import BoxOutputModel

#-----------------------------------
# Class
#-----------------------------------

class BoxCodeModel (AbstractBoxModel, AbstractCodeModel) :
    """Est le model des boxes dont le comportement est décrit par du code."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, name : str, path : str, creator : str, date : date, version : float, code : str = '', inputs : Iterable[BoxInputModel] = [], outputs : Iterable[BoxOutputModel] = []) : 
        
        AbstractBoxModel.__init__(self, name, path, creator, date, version, inputs, outputs)
        AbstractCodeModel.__init__(self, code)


#-----------------------------------
# Imports
#-----------------------------------

from typing import Dict

from .box_library_data import AbstractBoxData

#-----------------------------------
# Class
#-----------------------------------

class AbstractBoxLibraryImporter :
    """Est la classe abstraite commune permettant de charger des boxes."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self) :
        ...

    # -------------
    # Methods
    # -------------

    def import_box (self, path : str, file_name : str, absolute_yaml_file : str, boxes : Dict[str, AbstractBoxData]) -> Dict[str, AbstractBoxData]:
        """Est la méthode pour charger des boxes depuis des fichiers."""
        raise(NotImplementedError(f"{type(self).__name__}.import_box() not implemented yet"))


#-----------------------------------
# Imports
#-----------------------------------

import ctypes
import logging
import sys
import os
import traceback
from types import TracebackType
from time import sleep
from multiprocessing import Process, Manager
from threading import Thread

from PyQt5.QtGui import QIcon
from PyQt5.QtCore import QSize
from PyQt5.QtWidgets import QApplication

from muphyn.packages.core.box_library.plci_core_boxes_libraries import BoxesLibrariesManager
from muphyn.packages.interface.main_window import MainWindow 
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.core.utils.directory_manager import DirectoryManager
from muphyn.packages.core.utils.global_env_variables_manager import GlobalEnvVariablesManager

from muphyn.utils.paths import ROOT_DIR

from IPython.core.magic import Magics, line_magic, magics_class
from IPython import get_ipython

def muphyn_excepthook(exc_type: type[BaseException], exc_value: BaseException, exc_tb: TracebackType):
    # Save openned project as 
    if 'win' in globals():
        global win
        win.backup_save_projects()

    # Format Exception error 
    formatted_exception = \
        f"Error while running Muphyn: \n{''.join(traceback.format_exception(exc_type, exc_value, exc_tb))}"

    # Write Error in Log File
    LogManager().error(formatted_exception)

    # Quit Applcation
    QApplication.exit(-1)

#-----------------------------------
# Main method
#-----------------------------------
def run_muphyn(globals_dict = {}):
    #_ = os.system('cls')
    myappid = 'ceref.muphyn.1.2'

    if os.name == 'nt' :
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)

    # Replace excepthook function
    sys.excepthook = muphyn_excepthook

    # Init Managers
    
    LogManager().info("Starting MuPhyN application", is_global_message=True)
    GlobalEnvVariablesManager(global_vars=globals_dict)
    DirectoryManager()
    BoxesLibrariesManager()

    app = QApplication(sys.argv)

    app_icon = QIcon()
    app_icon.addFile(ROOT_DIR + "/" + "packages/interface/assets/MuPhyN.ico", QSize(400, 400))
    app.setWindowIcon(app_icon)

    # Load Style Sheet
    with open(ROOT_DIR + "/" + "themes/light.qss", "r") as styleSheetFile:
        styleSheetFileStr = styleSheetFile.read().replace("%ROOT_DIR%", ROOT_DIR).replace("\\", "/")
        app.setStyleSheet(styleSheetFileStr)

    win = MainWindow()
    win.show()

    # Launch Muphyn application
    exit_code = app.exec_()

    # Print in log the exit code
    LogManager().set_current_project_id(-1)
    LogManager().info(f"Closing MuPhyN application with code: {exit_code}", is_global_message=True)

    # Closing application
    sys.exit(exit_code)

def filter_local_variables(_global_vars = {}):
    _output = {}

    _excl_list = ['In', 'Out', 'Process', 'Thread', 'TracebackType']

    _type_incl_list = [str, int, float, list, dict, type]
    
    for _x in [key for key in _global_vars if not key.startswith('_')]:
        if type(_global_vars[_x]) in _type_incl_list and _x not in _excl_list:
            _output.update({_x:_global_vars[_x]})

    return _output

def update_globals(vars, ts = 0.25, linked_process:Process = None):
    if linked_process is None:
        return
    while(linked_process.is_alive()):
        sleep(ts)
        vars.clear()
        vars.update(filter_local_variables(globals().copy()))
    LogManager().info("Closing globals synchronization thread", is_global_message=True)

def start(data = {}):
    LogManager(log_level=logging.DEBUG)
    _globals = {}
    
    if len(data) == 0:
        _globals = Manager().dict()
        _globals.update(filter_local_variables(globals().copy()))
        LogManager().info(f"Getting data from globals: {_globals}", is_global_message=True)
        
    else:
        _globals.update(filter_local_variables(data))
    

    # TODO: automatically start MuPhyN by opening a file with the same name as the ipynb file 
    # using the key ['__vsc_ipynb_file__'] in globals() 
    # If no file exists, create it

    LogManager().info("Starting MuPhyN process", is_global_message=True)
    _muphyn_process = Process(target=run_muphyn, args=(_globals,))
    _muphyn_process.start()

    LogManager().info("Creating globals synchronization thread", is_global_message=True)
    update_thread = Thread(target=update_globals, args=(_globals, 0.25, _muphyn_process))
    update_thread.start()

if __name__ == '__main__':
    start()

@magics_class
class MuphynMagics(Magics):

    def __init__(self, shell):
        super(MuphynMagics, self).__init__(shell)
        self.default_runner = self.shell.safe_execfile

    @line_magic
    def muphyn(self, line):
        get_ipython().magic("run -i \"" + ROOT_DIR + "/"+ "runner.py\"")
    
    @line_magic
    def launch_muphyn(self, line):
        # TODO: find a way to launch the simulation from a magic command
        return


def load_ipython_extension(ipython):
    """ Enables to load the module with %load_ext muphyn """
    ipython.register_magics(MuphynMagics)


#-----------------------------------
# Imports
#-----------------------------------

#-----------------------------------
# Class
#-----------------------------------

class SchedulerParams :
    """Est la classe qui permet de maintenir les paramètres de la simulation dans une classe."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, stop_time_ : float = 10.0, step_time_ : float = 0.1) :
        self._stop_time = stop_time_
        self._step_time = step_time_

    # -------------
    # Properties
    # -------------

    @property 
    def step_time (self) -> float : 
        """Permet de récuperer le pas de temps pour la simulation."""
        return self._step_time

    @property 
    def stop_time (self) -> float :
        """Permet de récuperer le temps de fin de simulation."""
        return self._stop_time



#-----------------------------------
# Imports
#-----------------------------------

from typing import Any

from muphyn.packages.core.plci_core_scheduler import Scheduler

#-----------------------------------
# Methods
#-----------------------------------

def _execute_box (scheduler: Scheduler, event, box) -> None :
    
    #if event.signal is None :
    #    print("Running none event to ", box)
    #else : 
    #   print("Running event from ", event.box, " on ", event.signal, " to ", box)

    # print(box.name)
    
    new_events = box.function(event)

    if scheduler.is_scheduler_exception(new_events) :
        return new_events

    for event in new_events :
        
        # print("add event from ", event.box, " on ", event.signal)
        scheduler.append_event(event)

        if event.signal in scheduler.diagram.linked_signals :
            #print("event on ", event.signal, " at time", scheduler.current_time," and it is linked | # of events : ", scheduler.events_count)

            for signal in scheduler.diagram.linked_signals[event.signal] :
                scheduler.append_event(scheduler.construct_signal_event(signal, event.box, event.new_signal_data))
                #print(event.signal, " is connected to ", signal.index," | # of events now : ", scheduler.events_count)

def _default_scheduler_method (scheduler: Scheduler) -> Any :
    
    b_bis = None
    event = None
    b = None

    try :

        while scheduler.should_simulation_continue :
            
            for b in scheduler.diagram.boxes :
                
                new_events = _execute_box(scheduler, scheduler.construct_event(), b)
                if scheduler.is_scheduler_exception(new_events) :
                    return new_events

                while scheduler.are_events_left :
                    
                    event = scheduler.get_event(0)
                    event.signal.data = event.new_signal_data

                    if event.signal in scheduler.diagram.box_inputs :
                    
                        for b_bis in scheduler.diagram.box_inputs[event.signal] :
                            
                            new_events = _execute_box(scheduler, scheduler.construct_event(event), b_bis)
                            if scheduler.is_scheduler_exception(new_events) :
                                return new_events
                    
                    scheduler.remove_event(event)

            scheduler.stepping_time()

    except Exception as err :
        return scheduler.construct_exception(b, b_bis, event, err)


#-----------------------------------
# Imports
#-----------------------------------

# General Imports
import ctypes, yaml
from typing import Iterable, List

# PyQt5 Imports
from PyQt5 import QtGui
from PyQt5.QtCore import Qt, pyqtSignal, QPointF
from PyQt5.QtGui import QPainter, QIcon
from PyQt5.QtWidgets import QGraphicsScene, QGraphicsSceneDragDropEvent, QGraphicsView, QTabWidget, QApplication, QMenu, QAction
from muphyn.packages.core.simulation.model_parser.open_modelica_model_parser import OpenModelicaModelParser

# Project Imports
from muphyn.packages.interface.editors.abstract_editor import AbstractEditor
from muphyn.packages.interface.graphical_actions.actions_holder import ActionsHolder
from muphyn.packages.interface.graphical_actions.diagram_add_graphical_element_action import DiagramAddGraphicalElementAction
from muphyn.packages.interface.graphical_actions.diagram_paste_graphical_element_action import DiagramPasteGraphicalElementAction
from muphyn.packages.interface.graphical_actions.diagram_remove_graphical_element_action import DiagramRemoveGraphicalElementAction
from muphyn.packages.interface.models.editable_models.abstract_diagram_model import AbstractDiagramModel
from muphyn.packages.interface.models.graphical_models.abstract_box_model import AbstractBoxModel
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement
from muphyn.packages.core.box_library.box_library_data import AbstractBoxData
from muphyn.packages.interface.models.graphical_models.box_model import BoxModel
from muphyn.packages.interface.models.graphical_models.resizers.abstract_resizer import AbstractResizer
from muphyn.packages.interface.models.links_model.abstract_link_model import AbstractLinkModel
from muphyn.packages.interface.models.signals_model.input_connection_model import InputConnectionModel
from muphyn.packages.interface.models.signals_model.output_connection_model import OutputConnectionModel
import muphyn.packages.interface.files.simulation_files.simulation_exporter as exporter

from muphyn.packages.interface.widgets.multiphysics_model_previewer import MultiphysicsModelPreviewer
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.interface.widgets.StandardActions import RotateLeftAction, RotateRightAction, DeleteAction

#-----------------------------------
# Classes
#-----------------------------------

class SceneEditor (QGraphicsScene) :
    """Est la scéne graphique dans laquelle les éléments sont dessinés."""    

    # -------------
    # Signals
    # -------------
    
    elements_selected_changed = pyqtSignal(object)
    
    # -------------
    # Constructors
    # -------------

    def __init__ (self, diagram_editor : QGraphicsView, actions_holder : ActionsHolder, diagram_model : AbstractDiagramModel) :

        QGraphicsScene.__init__(self, diagram_editor)
        self._diagram_editor = diagram_editor
        self._actions_holder = actions_holder
        self._diagram_model = diagram_model
        self.setParent(diagram_editor)

    # -------------
    # Properties
    # -------------

    @property
    def actions_holder (self) -> ActionsHolder :
        """Permet de récuperer le conteneur d'action."""
        return self._actions_holder

    @property
    def diagram_model (self) -> AbstractDiagramModel :
        """Permet de récuperer le modèle content diagramme en cours d'édition."""
        return self._diagram_model

    # -------------
    # Methods
    # -------------

    def dropEvent (self, event : QGraphicsSceneDragDropEvent) -> None :
        if event.mimeData().hasUrls() :
            parent = self.parent()
            next_parent = parent.parent()

            while not(next_parent is None) :
                parent = next_parent
                next_parent = parent.parent()

            for url in event.mimeData().urls() :
                self._diagram_editor.open_file(url.toLocalFile())

        else :
            if event.possibleActions() == Qt.CopyAction and event.mimeData().data('action') == 'new box' :
                
                box_data : AbstractBoxData = ctypes.cast(int(event.mimeData().data('box_data')), ctypes.py_object).value

                offset_x, offset_y = [float(offset) for offset in event.mimeData().data('offset').split(";")]

                new_box_position = event.scenePos() + QPointF(offset_x, offset_y)

                action = DiagramAddGraphicalElementAction(self._diagram_editor._diagram_model, [{'box_data': box_data, 'pos': new_box_position}])
                action.do()
                self.actions_holder.append(action)
        
                event.accept()

            else :
                for graphical_element in self.parent()._diagram_model.graphical_elements :
                    if hasattr(graphical_element, 'inputs') :
                        for input in graphical_element.inputs :
                            if input.isUnderMouse() :
                                input.dropEvent(event)
                                if event.isAccepted() :
                                    return

        super().dropEvent(event)

    def dragEnterEvent (self, event : QGraphicsSceneDragDropEvent) -> None :
        
        if event.mimeData().hasUrls() :
            event.accept()
        
        else :

            if event.possibleActions() == Qt.CopyAction and event.mimeData().data('action') == 'new box' :
                    event.accept()
            
            if not(event.isAccepted()) : 
                for graphical_element in self.parent()._diagram_model.graphical_elements :
                    if hasattr(graphical_element, 'inputs') :
                        for input in graphical_element.inputs :
                            if input.isUnderMouse() :
                                input.dragEnterEvent(event)
                                if event.isAccepted() :
                                    return
            
        super().dragEnterEvent(event)
            
    
    def dragMoveEvent (self, event : QGraphicsSceneDragDropEvent) -> None :
        
        if event.mimeData().hasUrls() :
            event.accept()

        else :
            if event.possibleActions() == Qt.CopyAction and event.mimeData().data('action') == 'new box' :
                event.accept()
            else :
                for graphical_element in self.parent()._diagram_model.graphical_elements :
                    if isinstance(graphical_element, AbstractBoxModel) :
                        if hasattr(graphical_element, 'inputs') :
                            for input in graphical_element.inputs :
                                if input.isUnderMouse() :
                                    input.dragMoveEvent(event)
                                    if event.isAccepted() :
                                        return

        super().dragEnterEvent(event)

    def selected_elements (self) -> Iterable :
        """Permet de récuperer les éléments sélectionnés dans l'interface."""

        for item in self.selectedItems() :
            if isinstance(item, AbstractGraphicalElement) :
                if isinstance(item, AbstractResizer) :
                    continue

                yield item


class GraphicsView (QGraphicsView) :
    """Est la vue graphique dans laquelle la scène graphique est placée.""" 

    # -------------
    # Signals
    # -------------
    
    elements_selected_changed = pyqtSignal(object)

    # -------------
    # Constructors
    # -------------
    
    def __init__ (self, graphical_editor : AbstractEditor, actions_holder : ActionsHolder, diagram_model : AbstractDiagramModel) : 

        QGraphicsView.__init__(self, graphical_editor)

        self._diagram_model : AbstractDiagramModel = diagram_model 
        self._graphical_editor = graphical_editor
        self._actions_holder = actions_holder

        self.setDragMode(QGraphicsView.ScrollHandDrag)
        self.setDragMode(QGraphicsView.RubberBandDrag)

        self._diagram_model.scene = SceneEditor(self, actions_holder, diagram_model)
        self._diagram_model.scene.selectionChanged.connect(self.scene_elements_selected_changed)
        
        self.setScene(self._diagram_model.scene)

        self.setAcceptDrops(True)
        self.setMouseTracking(True)
        self._sliding_mouse_event_started =  False

        self.setRenderHint(QPainter.RenderHint.Antialiasing, on=True)
        self.setRenderHint(QPainter.RenderHint.TextAntialiasing, on=True)

    # -------------
    # Properties
    # -------------

    @property
    def actions_holder (self) -> ActionsHolder :
        """Permet de récuperer le conteneur d'action."""
        return self._actions_holder

    @property
    def diagram_model (self) -> AbstractDiagramModel :
        """Permet de récuperer le modèle content diagramme en cours d'édition."""
        return self._diagram_model

    # -------------
    # Methods
    # -------------
    def add_graphical_element (self, graphical_element : AbstractGraphicalElement) -> None :
        """Permet d'ajouter un élément graphique à l'interface."""

        if graphical_element is None :
            return

        self._diagram_model.add_element(graphical_element)

    def rem_graphical_element (self, graphical_element : AbstractGraphicalElement) -> None :
        """Permet de supprimer un élément graphique de l'interface."""

        self._diagram_model.remove_element(graphical_element)
        graphical_element.deleteLater()

    
    def mousePressEvent (self, event : QtGui.QMouseEvent):
        """Est l'événement appelée lorsque l'utilisateur appuie sur un bouton de sa souris."""


        if event.button() == Qt.MidButton :
            self.__prevMousePos = event.pos()
            self._sliding_mouse_event_started = True
            event.accept()

        elif event.button() == Qt.RightButton:
            while QApplication.overrideCursor() is not None:
                QApplication.restoreOverrideCursor()
            graphical_element_under_mouse = None
            for graphical_element in self.parent()._diagram_model.graphical_elements :
                if graphical_element.isUnderMouse():
                    graphical_element_under_mouse = graphical_element
                    break

            # If Right Click on Box
            if graphical_element_under_mouse is not None:
                # Init Context Menu Object
                contextMenu = QMenu()

                # Init possible actions
                rotate_left = QAction()
                rotate_right = QAction()
                delete_item = QAction()

                # Add actions if right-click on box
                if isinstance(graphical_element_under_mouse, AbstractBoxModel):
                    # Load Action
                    rotate_left = RotateLeftAction()
                    rotate_right = RotateRightAction()
                    delete_item = DeleteAction("Delete Box")

                    # Append Actions
                    contextMenu.addActions([rotate_left, rotate_right, delete_item])
                    
                # Add actions if right-click on link
                elif isinstance(graphical_element_under_mouse, AbstractLinkModel):
                    # Load Action
                    delete_item = DeleteAction("Delete Link")

                    # Append Actions
                    contextMenu.addActions([delete_item])

                # Add Separator
                contextMenu.addSeparator()

                #
                open_model = 0
                if isinstance(graphical_element_under_mouse, BoxModel):
                    if graphical_element_under_mouse.box_type == "multiphysics-simulation":
                        # Open Modelica
                        open_model = contextMenu.addAction("Open Modelica Model")
                        contextMenu.addSeparator()

                # General Actions
                remove_selected_elements = DeleteAction("Remove Selected Boxes")
                contextMenu.addActions([remove_selected_elements])

                action = contextMenu.exec_(event.screenPos().toPoint())
                
                if action == rotate_left:
                    new_rotation = (graphical_element_under_mouse.rotation() - 90) % 360
                    graphical_element_under_mouse.setRotation(new_rotation)
                elif action == rotate_right:
                    new_rotation = (graphical_element_under_mouse.rotation() + 90) % 360
                    graphical_element_under_mouse.setRotation(new_rotation)
                elif action == delete_item:
                    action = DiagramRemoveGraphicalElementAction(self._diagram_model, [graphical_element_under_mouse])
                    action.do()
                    self.actions_holder.append(action)
                elif action == remove_selected_elements:
                    action = DiagramRemoveGraphicalElementAction(self._diagram_model, self.selected_elements())
                    action.do()
                    self.actions_holder.append(action)
                elif action == open_model:
                    model_path = graphical_element_under_mouse._parameters["model_path"]["value"]
                    try:
                        # Open new tab with OM model
                        om_parser = OpenModelicaModelParser(model_path)
                        previewer = MultiphysicsModelPreviewer(om_parser, parent=self)
                        previewer.open()
                    except Exception as e:
                        LogManager().error(f"Can't open Open Modelica Previewer: {e}")
                else:
                    pass
            else:
                # Init Context Menu Object
                contextMenu = QMenu()

                # Init Action
                remove_selected_elements = DeleteAction("Remove Selected Boxes")

                # Add action to context menu
                contextMenu.addActions([remove_selected_elements])

                # Open Right-click menu
                action = contextMenu.exec_(event.screenPos().toPoint())
                
                if action == remove_selected_elements:
                    action = DiagramRemoveGraphicalElementAction(self._diagram_model, self.selected_elements())
                    action.do()
                    self.actions_holder.append(action)
                else:
                    pass

        else:
            super().mousePressEvent(event)

    def mouseMoveEvent (self, event : QtGui.QMouseEvent) :
        """Est l'événement appelée lorsque l'utilisateur bouge sa souris dans l'élément graphique."""
        
        if event.buttons() == Qt.MidButton :
            
            if not(self._sliding_mouse_event_started) :
                self.__prevMousePos = event.pos()
                self._sliding_mouse_event_started = True
            
            else :
                offset = self.__prevMousePos - event.pos()
                self.__prevMousePos = event.pos()

                self.verticalScrollBar().setValue(self.verticalScrollBar().value() + offset.y())
                self.horizontalScrollBar().setValue(self.horizontalScrollBar().value() + offset.x())

            event.accept()

        else :
            
            return super().mouseMoveEvent(event)


    def mouseReleaseEvent (self, event: QtGui.QMouseEvent) -> None :
        """Est l'événement appelée lorsque l'utilisateur relache un bouton de sa souris."""

        if event.button() == Qt.MidButton :
            self._sliding_mouse_event_started = False
            event.accept()

        else :
            return super().mouseReleaseEvent(event)

    def wheelEvent (self, event: QtGui.QWheelEvent) -> None :
        """Est l'événement appelée lorsque l'utilisateur bouge sa souris dans l'élément graphique."""

        if event.modifiers() == Qt.ControlModifier :

            self.zoom(event.angleDelta().y() / 120)
            event.accept()

        else :
            super().wheelEvent(event) 

    def zoom (self, value : int) -> None :
        
        if value == 0 :
            value = 1

        elif value > 0 :
            value *= 1.125

        elif value < 0 :
            value *= -0.825
        
        self.scale(value, value)

    def scene_elements_selected_changed (self) -> None :
        """Est la méthode appelée lorsque l'utilisateur modifie la sélection d'éléments à l'écran."""
        self.elements_selected_changed.emit(self.selected_elements)
    
    def selected_elements (self) -> Iterable :
        """Permet de récuperer les éléments sélectionnés dans l'interface."""
        
        for item in self._diagram_model.scene.selected_elements() :
            yield item

class DiagramEditor (AbstractEditor) :
    """Est l'élément graphique capable d'éditer des simulation ou des boxes composites.""" 

    # -------------
    # Constructors
    # -------------

    def __init__ (self, tab_holder : QTabWidget, diagram_model : AbstractDiagramModel, actions_holder : ActionsHolder) :

        AbstractEditor.__init__(self, tab_holder, diagram_model, actions_holder)

        self._diagram_model = diagram_model
        self._graphics_view : GraphicsView = GraphicsView(self, self._actions_holder, diagram_model) 
        self._graphics_view.setGeometry(0, 0, self.width(), self.height())
        self._graphics_view.elements_selected_changed.connect(self.graphics_view_elements_selected_changed)
        self.setMouseTracking(True)

    # -------------
    # Properties
    # -------------

    @property
    def diagram_model (self) -> AbstractDiagramModel :
        """Permet de récuperer le modèle content diagramme en cours d'édition."""
        return self._diagram_model 
    
    # -------------
    # Methods
    # -------------

    def copy (self) -> None :
        
        serializer = {}
        serializer['MuPhyN'] = {}
        serializer['MuPhyN']['boxes'] = []
        serializer['MuPhyN']['signals'] = []

        box_copied = []
        for box_model in self.selected_elements() :
            if isinstance(box_model, BoxModel) :

                if not(box_model in box_copied) : 
                    box_copied.append(box_model)
                
        link_copied : List[AbstractLinkModel] = []
        for link_model in self.selected_elements() :
            if isinstance(link_model, AbstractLinkModel) :

                if link_model.input.parent() in box_copied and link_model.output.parent() in box_copied :
                
                    if link_model in link_copied :
                        continue

                    link_dict = exporter.export_signal(len(link_copied), link_model)
                    serializer['MuPhyN']['signals'].append(link_dict)
                    link_copied.append(link_model)

        for box_model in box_copied :

            box_dict = box_model.to_dict()

            # box_dict = exporter.export_box(box_model)

            # box_dict['inputs'] = [] 
            # for input_model in box_model.inputs : 
                
            #     input_dict = exporter.export_input(input_model)
            #     input_dict['signal_index'] = -1
            #     if len(input_model) > 0 :
            #         link : AbstractLinkModel = input_model._links[0]
            #         if link in link_copied :
            #             input_dict['signal_index'] = link_copied.index(link)

            #     box_dict['inputs'].append(input_dict)

            # box_dict['outputs'] = [] 
            # for output_model in box_model.outputs : 

            #     output_dict = exporter.export_output(output_model)
            #     output_dict['signal_indices'] = []

            #     for link in output_model._links :
            #         if link in link_copied :
            #             output_dict['signal_indices'].append(link_copied.index(link))

            #     box_dict['outputs'].append(output_dict)


            serializer['MuPhyN']['boxes'].append(box_dict)
        
        cb = QApplication.clipboard()
        cb.clear(mode = cb.Clipboard)
        cb.setText(yaml.dump(serializer).__str__())

    def cut (self) -> None :
        
        self.copy()
        self.delete_selection()

    def paste (self) -> None :

        cb = QApplication.clipboard()
        cb_text = cb.text().strip()

        if cb_text.__len__() == 0 :
            return

        action = DiagramPasteGraphicalElementAction(self, self.diagram_model, cb_text)
        action.do()
        self.actions_holder.append(action)

    def resizeEvent (self, event : QtGui.QResizeEvent) -> None :
        self._graphics_view.setGeometry(0, 0, self.width(), self.height())
        return super().resizeEvent(event)

    def graphics_view_elements_selected_changed (self, elements) -> None :
        """Est la méthode appelée lorsque l'utilisateur change la sélection."""
        self.elements_selected_changed.emit(elements)
    
    def selected_elements (self) -> Iterable :
        for item in self._graphics_view.selected_elements() :
            yield item

    def unslect_elements (self) -> None :
        for el in self.selected_elements() :
            el.setSelected(False) 

    def elements (self) -> Iterable[AbstractGraphicalElement] :
        for chld in self._graphics_view._diagram_model._graphical_elements :
            if isinstance(chld, AbstractGraphicalElement) :

                if isinstance(chld, InputConnectionModel) or isinstance(chld, OutputConnectionModel) :
                    continue

                yield chld 

    def zoom (self, value : int) -> None :
        self._graphics_view.zoom(value)

    def add_item (self, graphical_element : AbstractGraphicalElement) -> None :
        self._graphics_view.add_graphical_element(graphical_element)
        
    def rem_item (self, graphical_element : AbstractGraphicalElement) -> None :
        self._graphics_view.rem_graphical_element(graphical_element)

    def delete_selection (self) -> None :
        action = DiagramRemoveGraphicalElementAction(self._graphics_view._diagram_model, self.selected_elements())
        action.do()
        self.actions_holder.append(action)

    def clear (self) -> None :
        to_delete = []
        for el in self.elements() :
            if isinstance(el, AbstractLinkModel) :
                continue
            to_delete.append(el)

        for el in to_delete :
            self.rem_item(el)


#-----------------------------------
# Imports
#-----------------------------------

from .plci_core_data import Data
from .plci_core_signal import Signal
from typing import Any

#-----------------------------------
# Class
#-----------------------------------

class SignalEvent :
    """Est la classe qui décrit le contenu d'un événement triggé par les boxes pour modifier les valeur des signaux."""

    # -------------
    # Constructors
    # -------------
    def __init__ (self, signal_ : Signal, box_ : Any, new_signal_data : Data) :
        self._signal = signal_
        self._box = box_
        self._last_signal_data = signal_.data
        self._new_signal_data = new_signal_data

    # -------------
    # Properties
    # -------------

    @property 
    def signal (self) -> Signal : 
        """Permet de récuperer le signal de l'événement."""
        return self._signal

    @property
    def box (self) -> Any :
        """Permet de récuperer la box de l'événement."""
        return self._box

    @property
    def last_signal_data (self) -> Data :
        """Permet de récuperer la donnée qui se trouvait sur le signal avant le changement."""
        return self._last_signal_data

    @property
    def last_signal_value (self) -> Data :
        """Permet de récuperer la valeur qui se trouvait sur le signal avant le changement."""
        return self._last_signal_data.value

    @property
    def new_signal_data (self) -> Data :
        """Permet de récuperer la nouvelle donnée."""
        return self._new_signal_data
    
    @property
    def new_signal_value (self) -> Any :
        """Permet de récuperer la nouvelle valeur."""
        return self._new_signal_data.value


from typing import Iterable
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QFileDialog, QHBoxLayout, QLineEdit

from muphyn.packages.interface.properties_pages.data_property_widgets.abstract_property_widget import AbstractPropertyWidget
from muphyn.packages.interface.utils.graphical_items.file_selector_button import file_selector_button, ExistingFilesSelectorButton

class AbstractPathPropertyWidget(AbstractPropertyWidget):

    def __init__(self, file_mode: QFileDialog.FileMode, parent = None, flags = Qt.Widget) -> None:
        super().__init__(parent, flags)
        
        # Init VBoxLayout
        layout = QHBoxLayout()

        # Add text field
        self.textfield = QLineEdit()
        self.textfield.editingFinished.connect(self.on_textfield_path_edited)

        # Add button to explore file
        self.select_path_button = file_selector_button(file_mode)
        self.select_path_button.accepted.connect(self.on_path_selector_accepted)

        # Set Layout
        layout.addWidget(self.textfield, 1)
        layout.addWidget(self.select_path_button)
        self.setLayout(layout)

    # -------------
    # Properties
    # -------------

    # -------------
    # Methods
    # -------------
    def on_path_selector_accepted(self):
        # Change path value
        self.setValue(self.select_path_button.path)

    def on_textfield_path_edited(self):
        # change path value
        self.setValue(self.textfield.text())

    def setValue(self, new_value: str):
        self._value = new_value

        # Update LineEdit
        if self.textfield.text() != new_value:
            self.textfield.setText(new_value)

        # Emit path changed
        self.valueChanged.emit()
        

class AnyFilePropertyWidget(AbstractPathPropertyWidget):
    def __init__(self, parent=None) -> None:
        super().__init__(QFileDialog.FileMode.AnyFile, parent)        

class DirectoryPropertyWidget(AbstractPathPropertyWidget):
    def __init__(self, parent=None) -> None:
        super().__init__(QFileDialog.FileMode.Directory, parent)

class ExistingFilePropertyWidget(AbstractPathPropertyWidget):
    def __init__(self, parent=None) -> None:
        super().__init__(QFileDialog.FileMode.ExistingFile, parent)


class ExistingFilesPropertyWidget(AbstractPropertyWidget):

    def __init__(self, parent = None, flags = Qt.Widget) -> None:
        super().__init__(parent, flags)

        # Path
        self._value = ""

        # Init VBoxLayout
        layout = QHBoxLayout()
        layout.setAlignment(Qt.AlignmentFlag.AlignTop)

        # Add button to explore file
        self.select_path_button = ExistingFilesSelectorButton()
        self.select_path_button.accepted.connect(self.on_path_selector_accepted)

        # Set Layout
        layout.addWidget(self.select_path_button)
        self.setLayout(layout)

    # -------------
    # Properties
    # -------------

    # -------------
    # Methods
    # -------------
    def on_path_selector_accepted(self):
        # Change path value
        self.setValue(self.select_path_button.path)

    def setValue(self, new_value: Iterable[str]):
        self._value = new_value

        # Emit path changed
        self.valueChanged.emit()



#-----------------------------------
# Imports
#-----------------------------------

from typing import Any

#-----------------------------------
# Class
#-----------------------------------

class EventSignalData :
    """Est la classe des événements véhiculés par les resizers."""

    # -------------
    # Contructors
    # -------------

    def __init__ (self, sender : Any, value : Any) :
        self._sender : Any = sender
        self._value : Any = value

    # -------------
    # Properties
    # -------------

    @property
    def sender (self) -> Any :
        return self._sender

    @property
    def value (self) -> Any :
        return self._value


#-----------------------------------
# Imports
#-----------------------------------

from typing import List
from random import seed
from random import random

#-----------------------------------
# Functions
#-----------------------------------

def _init_noise_box (box, simulation_params) -> List : 
    
    if box['apply_seed'] :
        seed(box['seed'])

    box['min'] = box['mean_value'] - (box['amplitude'] / 2)

def _function_noise_box (box, event_) -> List :

    v = box['amplitude'] * random() + box['min']
    
    events : List = []
    
    for output in box.outputs :
        events.append(box.construct_signal_event(output, v))

    return events


#-----------------------------------
# Imports
#-----------------------------------

# PyQt5 Imports
from PyQt5.QtCore import QObject, QPointF, QSizeF, pyqtSignal
from PyQt5.QtWidgets import QGraphicsItem, QGraphicsSceneMouseEvent

# Project Imports
from muphyn.packages.interface.shapes.basic_shape.grouped_shapes import GroupedShapes

#-----------------------------------
# Class
#-----------------------------------
class AbstractGraphicalElement (GroupedShapes) :
    """Est la classe abstraite commune aux éléments graphiques affichés dans l'interface."""

    # -------------
    # Static Element
    # -------------

    GRAPHICAL_INDEX : int = 0
    
    # -------------
    # Signals
    # -------------

    param_changed = pyqtSignal(object, str, object, object)
    """
    Params : 
    - *object [AbstractGraphicalElement]* model
    - *str* param name
    - *object* old value
    - *object* new value
    """

    # -------------
    # Constructors
    # -------------

    def __init__ (self, name : str, position : QPointF, rotation : float = 0, text : str = '', parent : QGraphicsItem = None) :
        
        # Init Inherited class
        GroupedShapes.__init__(self, position, parent)

        # Set Graphical Index
        self._graphical_index = AbstractGraphicalElement.GRAPHICAL_INDEX
        AbstractGraphicalElement.GRAPHICAL_INDEX += 1

        # Init Action Semaphore
        self._action_size_semaphore : bool = True 
        self._action_pos_semaphore : bool = True
        self._action_rot_semaphore : bool = True
        self._action_param_semaphore : bool = True

        # Init Item Parameters
        self._name : str = name
        self._text = text

        self.setRotation(rotation)
        
    # -------------
    # Properties
    # -------------

    @property
    def graphical_index (self) -> int :
        """
        Permet de récuperer l'index graphique.
        Cet index est utilisé pour référencé les objets pour les actions.
        Il permet de retenir quel élément est quel élément peut importe l'ordre des graphical elements dans le diagram.
        Il n'est pas nécessaire de le sauvegarder dans un fichier. Les actions ne portent que sur les éléments actuellement dans l'interface.
        """
        return self._graphical_index

    @graphical_index.setter
    def graphical_index (self, new_graphical_index_ : int) -> None : 
        """Permet de modifier l'index graphique."""

        if new_graphical_index_ >= 0 and new_graphical_index_ != self._graphical_index :
            self._graphical_index = new_graphical_index_

    @property
    def name (self) -> str :
        """Permet de récuperer le nom de l'élément."""
        return self._name

    @name.setter
    def name (self, new_name : str) -> None :
        """Permet de modifier le nom de l'élément."""
        if new_name != self._name:
            self._name = new_name

    @property
    def position (self) -> QPointF :
        """Permet de récuperer la position de l'élément graphique."""
        return self.pos()

    @position.setter
    def position (self, new_position : QPointF) -> None :
        """Permet de modifier la position de l'élement graphique."""
        if new_position != self.pos():
            self.setPos(new_position) 

    @property
    def size (self) -> QSizeF :
        """Permet de récuperer la taille de l'élement graphique."""
        return self.boundingRect().size()

    @property
    def rot (self) -> float :
        """Permet de récuperer la rotation de la box."""
        return self.rotation()

    @rot.setter
    def rot (self, new_rot : float) -> None :
        """Permet de modifier la rotation de la box."""
        if new_rot != self.rotation():
            self.setRotation(new_rot)

    @property 
    def text (self) -> str :
        """Permet de récuperer le texte à afficher."""
        return self._text

    @text.setter
    def text (self, new_text: str) -> None :
        """Permet de modifier le texte à afficher."""

        if new_text != self._text:
            old_text = self._text
            self._text = new_text or '' 

            if self.action_param_semaphore :
                return

            self.param_changed.emit(self, 'text', old_text, self._text)
        
    @property
    def rendered_text (self) -> str :
        """Permet de récuperer le texte à afficher."""

        if self._text is None : 
            return self._name

        if self._text.strip().__len__() == 0 :
            return self._name

        return self._text

    @rendered_text.setter
    def rendered_text (self, new_rendered_text : str) -> None :
        """Permet de modifier le texte à afficher."""
        self.text = new_rendered_text

    @property
    def diagram_model (self) :
        """Permet de récuperer l'éditeur visuel."""
        return self._diagram_model

    @diagram_model.setter
    def diagram_model (self, diagram_model_) -> None :
        """Permet de modifier l'éditeur visuel."""
        self._diagram_model = diagram_model_

        if self._diagram_model is None :
            self._action_size_semaphore : bool = True 
            self._action_pos_semaphore : bool = True
            self._action_rot_semaphore : bool = True
            self._action_param_semaphore : bool = True
        
        else : 
            self._action_size_semaphore : bool = False 
            self._action_pos_semaphore : bool = False
            self._action_rot_semaphore : bool = False
            self._action_param_semaphore : bool = False

    @property
    def action_size_semaphore (self) -> bool :
        """Permet de récuperer l'état du sémpahore pour le changement de taille de l'objet."""
        return self._action_size_semaphore

    @action_size_semaphore.setter 
    def action_size_semaphore (self, action_size_semaphore_ : bool) :
        """Permet de modifier l'état du sémaphore pour le changement de taille de l'objet."""
        self._action_size_semaphore = action_size_semaphore_

    @property
    def action_pos_semaphore (self) -> bool :
        """Permet de récuperer l'état du sémpahore pour le changement de position de l'objet."""
        return self._action_pos_semaphore

    @action_pos_semaphore.setter 
    def action_pos_semaphore (self, action_pos_semaphore_ : bool) :
        """Permet de modifier l'état du sémaphore pour le changement de position de l'objet."""
        self._action_pos_semaphore = action_pos_semaphore_

    @property
    def action_rot_semaphore (self) -> bool :
        """Permet de récuperer l'état du sémpahore pour le changement de rotation de l'objet."""
        return self._action_rot_semaphore

    @action_rot_semaphore.setter 
    def action_rot_semaphore (self, action_rot_semaphore_ : bool) :
        """Permet de modifier l'état du sémaphore pour le changement de rotation de l'objet."""
        self._action_rot_semaphore = action_rot_semaphore_

    @property
    def action_param_semaphore (self) -> bool :
        """Permet de récuperer le sémaphore pour les paramètres de l'objet."""
        return self._action_param_semaphore

    @action_param_semaphore.setter
    def action_param_semaphore (self, action_param_semaphore_ : bool) -> None :
        """Permet de modifier le sémaphore pour les paramètres de l'objet."""
        self._action_param_semaphore = action_param_semaphore_

    # -------------
    # Methods
    # -------------
    def mouseDoubleClickEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        return super().mouseDoubleClickEvent(event)


    def parent(self) -> QObject:
        parent = super().parent()
        return parent if parent is not None else self.parentItem()


# -----------------------------------
# Imports
# -----------------------------------

from typing import List


# -----------------------------------
# Functions
# -----------------------------------

def euler(t_1, t_0, dt):
    """Permet de retourner la differentiel entre deux points."""
    return (t_1 - t_0) / dt

def get_coeff_vector(coeff_vector_string: str) -> List[float]:

    if coeff_vector_string is None:
        return [0]

    vector = []
    for sub in coeff_vector_string.split(' '):

        sub = sub.strip()
        if len(sub) == 0:
            continue

        try:
            vector.append(float(sub))
        except ValueError:
            ...

    return vector


def _init_transfert_function(box, simulation_params) -> None:
    
    box['num_vect'] = get_coeff_vector(box['numerator'])
    box['num_vect'].reverse()
    box['num_order'] = len(box['num_vect']) - 1
    box['last_u_derivatives'] = [0 for _ in box['num_vect']]

    box['denom_vect'] = get_coeff_vector(box['denominator'])
    box['denom_vect'].reverse()
    box['denom_order'] = len(box['denom_vect']) - 1
    box['last_y_derivatives'] = [0 for _ in box['denom_vect']]

    box['order'] = box['denom_order']
    box['order_coeff'] = box['denom_vect'][box['order']]

    box['last_y'] = box['initial_value']
    box['last_timing'] = -1

def _function_transfert_function(box, event_) -> List : 

    v = 0
    current_u_derivatives = []
    current_y_derivatives = []
    input_value = box.get_input(0).value

    if box['last_timing'] == event_.timing :
        return []
    box['last_timing'] = event_.timing

    for j, coeff_u in enumerate(box['num_vect']) : 

        if j == 0 : 
            v += coeff_u * input_value
            current_u_derivatives.append(input_value)
        
        else :
            derivative = (current_u_derivatives[j - 1] - box['last_u_derivatives'][j - 1]) / event_.step_time
            v += coeff_u * derivative
            current_u_derivatives.append(derivative)

    if box['order'] > 0 :

        for j in range(box['order']) :

            coeff_y = box['denom_vect'][j]

            if j == 0 :
                v -= coeff_y * box['last_y']
                current_y_derivatives.append(box['last_y'])

            else :
                derivative = (current_y_derivatives[j - 1] - box['last_y_derivatives'][j - 1]) / event_.step_time
                v -= coeff_y * derivative
                current_y_derivatives.append(derivative)

        v /= box['order_coeff']

        for j in range(box['order']).__reversed__() :
            # LogManager().debug(f'time : {event_.timing} - order : {j} - derivative : {current_y_derivatives[j]}')
            v *= event_.step_time
            v += current_y_derivatives[j]

    else :
        v /= box['order_coeff']

    box['last_u_derivatives'] = current_u_derivatives
    box['last_y_derivatives'] = current_y_derivatives
    box['last_y'] = v

    """
    while (len(box['last_y_s']) > box['denom_order']) :
        del box['last_y_s'][0]
    """


    events: List = []

    for output in box.outputs :
        events.append(box.construct_signal_event(output, v))

    return events



# -----------------------------------
# Imports
# -----------------------------------

from typing import List
from scipy.integrate import solve_ivp

from muphyn.packages.interface.models.graphical_models.box_model import BoxModel
from muphyn.packages.interface.utils.tools.math_expression_painter import mathTex_to_QPixmap

# -----------------------------------
# Functions
# -----------------------------------
def _update_icon(box_model: BoxModel):
    num = get_coeff_vector(box_model.get_parameter('numerator')["value"])
    num_str = ""
    for index in range(len(num)):
        degree = len(num) - index - 1
        coef = num [index]
        num_str += f"{' + ' if index != 0 else ''}{coef}" if coef >= 0 else f"-{-coef}"
        if degree > 1:
            num_str += f" * s^{degree}"
        elif degree == 1:
            num_str += " * s"

    denom = get_coeff_vector(box_model.get_parameter('denominator')["value"])
    denom_str = ""
    for index in range(len(denom)):
        degree = len(denom) - index - 1
        coef = denom [index]
        denom_str += f"{' + ' if index != 0 else ''}{coef}" if coef >= 0 else f"-{-coef}"
        if degree > 1:
            denom_str += f" * s^{degree}"
        elif degree == 1:
            denom_str += " * s"

    print("\\frac{" + num_str + "}{" + denom_str + "}")
    pixmap = mathTex_to_QPixmap("\\frac{" + num_str + "}{" + denom_str + "}")
    box_model.setIcon(pixmap)

def get_coeff_vector(coeff_vector_string: str) -> List[float]:

    if coeff_vector_string is None:
        return [0]

    vector = []
    for sub in coeff_vector_string.split(' '):

        sub = sub.strip()
        if len(sub) == 0:
            continue

        try:
            vector.append(float(sub))
        except ValueError:
            ...

    return vector


def _init_transfer_function(box, simulation_params) -> None:
    
    box['num_vect'] = get_coeff_vector(box['numerator'])
    box['num_vect'].reverse()
    box['num_order'] = len(box['num_vect']) - 1

    box['denom_vect'] = get_coeff_vector(box['denominator'])
    box['denom_vect'].reverse()
    box['denom_order'] = len(box['denom_vect']) - 1

    box['last_y'] = [box['initial_value']] + [0] * (box['denom_order'] -1) 
    
    box['last_timing'] = -1

def f(t, y, box): 
    out = 0
    for j, coeff_y  in enumerate(box['denom_vect'][:-1]):
        out -= coeff_y * y[j]
    
    for j, coeff_y in enumerate(box['num_vect']):
        if j == 0:
            out += coeff_y * box.get_input(0).value
        else:
            out += coeff_y * y[j]
    
    out_vector = [y[box['denom_order'] - 1 - x] for x in range(box['denom_order'])]
    out_vector[-1] = out/box['denom_vect'][-1]

    return out_vector

def _function_transfer_function(box, event_) -> List :             

    if box['last_timing'] == event_.timing:    # if we are still on the current step, leave
        return []

    events: List = []

    
    step = event_.step_time

    t0 = event_.timing

    t = [t0, t0 + step]                 # time vector is only current and next step
    sol = solve_ivp(f, t, box['last_y'], args=(box,), method='RK45', rtol=box['rtol'], atol=box['atol'])
    
    out = list(sol.y[:,-1])               # use the last values as the output of the step

    box['last_y'] = out

    for output in box.outputs :
        events.append(box.construct_signal_event(output, out[0])) # only send the y(t) output, discard eventual derivatives for order >= 2

    box['last_timing'] = event_.timing

    return events




from ..core.plci_core_data_type import DataType
from ..core.plci_core_signal import Signal
from ..core.plci_core_scheduler import Scheduler
from ..core.plci_core_scheduler_params import SchedulerParams

from .unit_test import get_boxes_libraries, get_schedulers_library, simulate

# Get library
boxesLibraries = get_boxes_libraries()
schedulersLibraries = get_schedulers_library()

print("Creation of the scheduler")
scheduler : Scheduler = schedulersLibraries.construct_scheduler('Schedulers', 'Default')
print('Schedulers libraries : ', schedulersLibraries)
print('Schedulers libraries count : ', schedulersLibraries.libraries.__len__())
print('Schedulers libraries[0] count : ', schedulersLibraries.libraries[0].schedulers.__len__())
print('scheduler : ', scheduler)
scheduler.params = SchedulerParams(5, 0.001)

print("Creation of the diagram")
# 0 Sinus
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sources', 'Sine', amplitude = 10, pulsation = 10, phase = 0)) 
# 1 Sinus
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sources', 'Sine', amplitude = 5, pulsation = 20, phase = 0))  
# 2 Addition
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Math', 'Addition'))                      
# 3 Graph box            
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sinks', 'Graph', start_time = 0, stop_time = scheduler.stop_time, title = 'Sinus', label_x = 'Temps (s)', label_y = 'Amplitude (v)' ))
# 4 Sinus
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sources', 'Sine', amplitude = 1, pulsation = 100, phase = 0))
# 5 Ramp
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sources', 'Ramp', start_time = 0, initial_value = 1, slope = -0.20))
# 6 Multiplication
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Math', 'Multiplier', start_time = 0, initial_value = 1, slope = -0.20))

print(scheduler.diagram._boxes[5])

# Diagram creation

scheduler.diagram.append(Signal(index_ = 0, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 1, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 2, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 3, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 4, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 5, signal_type_ = DataType.FLOAT, default_value_ = 0.0))

scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[0], scheduler.diagram.signals[0])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[1], scheduler.diagram.signals[1])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[2], scheduler.diagram.signals[2])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[4], scheduler.diagram.signals[3])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[5], scheduler.diagram.signals[4])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[6], scheduler.diagram.signals[5])

scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[2], scheduler.diagram.signals[0], scheduler.diagram.signals[1], scheduler.diagram.signals[3])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[6], scheduler.diagram.signals[2], scheduler.diagram.signals[4])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[3], scheduler.diagram.signals[5])

print("Boxes created : ")
for box in scheduler.diagram.boxes :
    print("\t - ", box) 

# Launch solver
simulate(scheduler)


# Component classes
from muphyn.packages.interface.editors.multiphysics_model.components.abstract_component import AbstractComponentModel, ComponentConnector
from muphyn.packages.interface.editors.multiphysics_model.components.electrical_components import ResistorModel, VariableResistorModel, GroundModel, SignalCurrentSourceModel

# Link
from muphyn.packages.interface.editors.multiphysics_model.connection_model import ConnectionModel

# Multiphysics model classes
from muphyn.packages.interface.editors.multiphysics_model.abstract_multiphysics_model import AbstractMultiPhysicsModel

# OpenModelica Model
from muphyn.packages.interface.editors.multiphysics_model.open_modelica_multiphysics_model import OpenModelicaMultiphysicsModel



#-----------------------------------
# Imports
#-----------------------------------

import locale
import csv
from typing import List

from muphyn.packages.core.utils.log_manager import LogManager

#-----------------------------------
# Functions
#-----------------------------------

def _init_csv_box (box, simulation_params) -> List :
    box['values'] = []
    box['point_count'] = 0

    titles = []
    titles.append('Time')
    v = 0
    for input in box.inputs :
        titles.append(str(v))
        v += 1
    box['values'].append(titles)

def _function_csv_box (box, event_) -> List :

    if box['copy_timing_from_simulation'] is True or (event_.timing > box['start_time'] and event_.timing < box['stop_time']) :
        lst = []
        lst.append(event_.timing)

        for input in box.inputs :
            lst.append(input.value)

        box['point_count'] = box['point_count'] + 1
        box['values'].append(lst)

    return []

def _end_csv_box (box) -> None : 
    LogManager().info('decimal delimiter : ', locale.localeconv()["decimal_point"])

    with open(box['file_name'], 'w', newline=box['new_line']) as csv_file : 
        writer = csv.writer(csv_file, delimiter=box['delimiter'], quotechar='\'', quoting=csv.QUOTE_MINIMAL)
        writer.writerows(box['values'])



#-----------------------------------
# Imports
#-----------------------------------

from PyQt5.QtGui import QColor
from PyQt5.QtCore import QPointF, QSizeF, Qt
from PyQt5.QtWidgets import QGraphicsItem

from muphyn.packages.core.plci_core_data_type import DataType

from muphyn.packages.interface.models.graphical_models.abstract_box_IO_model import AbstractBoxIOModel

from muphyn.packages.interface.models.signals_model.output_connection_model import OutputConnectionModel

from muphyn.packages.interface.models.signals_model.signal_link_model import SignalLinkModel

#-----------------------------------
# Class
#-----------------------------------

class BoxInputModel (AbstractBoxIOModel) :
    """Est la classe des entrées des boxes composite."""
    
    # -------------
    # Constructors
    # -------------

    def __init__ (self, name : str, data_type : DataType, position : QPointF, size : QSizeF,
                  rotation : float = 0.0, link : SignalLinkModel = None, text : str = '',
                  color: QColor = Qt.GlobalColor.black, parent : QGraphicsItem = None) :

        AbstractBoxIOModel.__init__(self, name, data_type, position, size, rotation, text, parent)
        
        self._outputs.append(OutputConnectionModel('', data_type, QPointF(0, 0), link, None, '', color, self))



#-----------------------------------
# Imports
#-----------------------------------

from typing import List
import math

#-----------------------------------
# Functions
#-----------------------------------

def _init_square_box_1 (box, simulation_params) -> None :
    
    box['current_period'] = 0
    box['period_max'] = (math.pi * 2) / box['pulsation']
    box['switch_period'] = box['period_max'] * box['duty_cycle']
    box['high_value'] = box['mean_value'] + (box['amplitude'] / 2)
    box['low_value'] = box['mean_value'] - (box['amplitude'] / 2)


def _function_square_box_1 (box, event_) -> List : 

    if event_.timing == 0 :
        box['current_period'] = 0
    else :
        box['current_period'] = box['current_period'] + event_.step_time

    while box['current_period'] >= box['period_max'] :
        box['current_period'] = box['current_period'] - box['period_max']
        
    if box['current_period'] < box['switch_period'] :
        v = box['low_value']

    elif box['current_period'] >= box['switch_period'] :
        v = box['high_value']

    events : List = []
    
    for output in box.outputs :
        events.append(box.construct_signal_event(output, v))

    return events


from PyQt5.QtCore import QPointF, QRectF, pyqtSignal
from PyQt5.QtGui import QPainter
from PyQt5.QtWidgets import QGraphicsItem, QGraphicsObject, QStyleOptionGraphicsItem

class GroupedShapes(QGraphicsObject):

    position_changed = pyqtSignal()

    def __init__(self, group_position: QPointF = QPointF(0, 0), parent: QGraphicsItem = None) -> None:
        QGraphicsObject.__init__(self, parent)        
        self.setPos(group_position)
    
    # -------------
    # Methods
    # -------------
    def boundingRect(self) -> QRectF:
        return self.childrenBoundingRect()

    def paint (self, painter: QPainter, option: QStyleOptionGraphicsItem = QStyleOptionGraphicsItem.OptionType.SO_Frame, widget=None) -> None :
        return None

    def setRotation(self, angle: float) -> None:
        self.setTransformOriginPoint(self.boundingRect().center())
        return super().setRotation(angle)


from typing import Iterable
from PyQt5.QtCore import QEventLoop, pyqtSignal
from PyQt5.QtWidgets import QFileDialog, QPushButton

from muphyn.packages.core.utils.log_manager import LogManager

class AbstractFileDialogButton(QPushButton):

    accepted = pyqtSignal()
    rejected = pyqtSignal()

    SupportedFileModes = [
        QFileDialog.FileMode.AnyFile,
        QFileDialog.FileMode.Directory,
        QFileDialog.FileMode.ExistingFile,
        QFileDialog.FileMode.ExistingFiles,
    ]

    def __init__(self, file_mode: QFileDialog.FileMode, parent = None):
        super().__init__("Ouvrir", parent)

        # Init path value
        self.path = None

        # FileMode
        self.file_mode = file_mode

        # Connect button
        self.clicked.connect(self.open_file_dialog)

    def open_file_dialog(self):
        # Open a file dialog
        self.file_dialog = QFileDialog(self)
        self.file_dialog.setFileMode(self.file_mode)

        # Open file dialog
        self.file_dialog.open()

        # Create wait for file selector blocking loop
        wait_for_file_selected_loop = QEventLoop()
        # Accepted event
        self.file_dialog.accepted.connect(self.handle_accepted_event)
        self.file_dialog.accepted.connect(wait_for_file_selected_loop.exit)

        # Rejected event
        self.file_dialog.rejected.connect(self.handle_rejected_event)
        self.file_dialog.rejected.connect(wait_for_file_selected_loop.exit)
        wait_for_file_selected_loop.exec()

    def handle_accepted_event(self):
        self.handle_selected_files(self.file_dialog.selectedFiles())

    def handle_rejected_event(self):
        self.rejected.emit()

    def handle_selected_files(self, selected_files: Iterable[str]) -> None:
        raise(NotImplementedError(f"{type(self).__name__}.handle_selected_files() not implemented yet"))

class AnyFileSelectorButton(AbstractFileDialogButton):

    def __init__(self, parent = None):
        super().__init__(QFileDialog.FileMode.AnyFile, parent)

    def handle_selected_files(self, selected_files: Iterable[str]) -> None:
                
        # SelectedFiles must contain only one path
        if len(selected_files) == 1: 
            # Change path value
            self.path = selected_files[0]

            # Emit file path has changed
            self.accepted.emit()
        else:
            self.rejected.emit()

class DirectorySelectorButton(AbstractFileDialogButton):

    def __init__(self, parent = None):
        super().__init__(QFileDialog.FileMode.Directory, parent)

    def handle_selected_files(self, selected_files: Iterable[str]) -> None:
                
        # SelectedFiles must contain only one directory
        if len(selected_files) == 1: 
            # Change path value
            self.path = selected_files[0]

            # Emit file path has changed
            self.accepted.emit()
        else:
            self.rejected.emit()

class ExistingFileSelectorButton(AbstractFileDialogButton):

    def __init__(self, parent=None):
        super().__init__(QFileDialog.FileMode.ExistingFile, parent)

    def handle_selected_files(self, selected_files: Iterable[str]) -> None:

        # SelectedFiles must contain only one file
        if len(selected_files) == 1: 
            # Change path value
            self.path = selected_files[0]

            # Emit accepted
            self.accepted.emit()
        else:
            # Emit rejected
            self.rejected.emit()

class ExistingFilesSelectorButton(AbstractFileDialogButton):

    def __init__(self, parent=None):
        super().__init__(QFileDialog.FileMode.ExistingFiles, parent)

    def handle_selected_files(self, selected_files: Iterable[str]) -> None:

        # SelectedFiles must contain at least one file
        if len(selected_files) > 0: 
            # Change path value
            self.path = selected_files

            # Emit accepted
            self.accepted.emit()
        else:
            # Emit rejected
            self.rejected.emit()

# Factory
def file_selector_button(file_mode: QFileDialog.FileMode, parent=None) -> AbstractFileDialogButton:
    if file_mode == QFileDialog.FileMode.AnyFile:
        return AnyFileSelectorButton(parent)
    elif file_mode == QFileDialog.FileMode.Directory:
        return DirectorySelectorButton(parent)
    elif file_mode == QFileDialog.FileMode.ExistingFile:
        return ExistingFileSelectorButton(parent)
    elif file_mode == QFileDialog.FileMode.ExistingFiles:
        return ExistingFilesSelectorButton(parent)
    else:
        supported_file_modes = '\n - '.join(AbstractFileDialogButton.SupportedFileModes)
        LogManager().error(f"Unsupported QFileDialog.FileMode value given: {file_mode}.\nList of supported QFileDialog.FileMode:\n - {supported_file_modes}")



from PyQt5.QtWidgets import QLabel, QTabWidget
from muphyn.packages.interface.editors.abstract_editor import AbstractEditor
from muphyn.packages.interface.graphical_actions.actions_holder import ActionsHolder
from muphyn.packages.interface.models.editable_models.abstract_code_model import AbstractCodeModel


class CodeEditor (AbstractEditor) :
    
    def __init__(self, tab_holder : QTabWidget, code_model : AbstractCodeModel, actions_holder : ActionsHolder) :
        AbstractEditor.__init__(self, tab_holder, code_model, actions_holder)
        
        self.lbl : QLabel = QLabel(self)
        self.lbl.setText('code editor')